#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class FolderError(Exception):
	'''
	Base class for exceptions occurring in the Folder.
	'''

	pass

class ConfigNotFoundError(FolderError):
	'''
	Exception raised when we try to access a non-existing configuration folder.

	Parameters
	----------
	foldername : str
		Name of the folder.
	'''

	def __init__(self, foldername):
		self.foldername = foldername

class NoConfigError(FolderError):
	'''
	Exception raised when we try to access a configuration but no foldername is given.
	'''

	pass

class SkeletonsNotFoundError(FolderError):
	'''
	Exception raised when we try to access a non-existing skeletons folder.

	Parameters
	----------
	skeletons_name : str
		Name of the folder.
	'''

	def __init__(self, skeletons_name):
		self.skeletons_name = skeletons_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .item import UIDisplayedItem

class UIProgressBar(UIDisplayedItem):
	'''
	Represent a progress bar displayed in the UI.

	Parameters
	----------
	ui : UI
		The UI object this text line belongs to.

	total : int
		The final number to reach.

	bar_length : int
		Length of the progress bar.

	empty_char : str
		Character to use for the empty part of the bar.

	full_char : str
		Character to use to fill the bar.

	percentage_precision : int|str
		Precision to use for the display of the percentage.
		Special value `'auto'` to guess the needed precision from the total.
	'''

	def __init__(self, ui, total, *, bar_length = 40, empty_char = '░', full_char = '█', percentage_precision = 'auto'):
		super().__init__(ui)

		self._total = total
		self._counter = 0

		self._bar_length = bar_length
		self._empty_char = empty_char
		self._full_char = full_char

		if percentage_precision == 'auto':
			self._percentage_precision = abs(floor(log10(100 / self._total))) if self._total > 100 else 0

		else:
			self._percentage_precision = percentage_precision

		self._pattern = ' '.join([
			f'{{counter:>{len(str(self._total))}d}}/{self._total}',
			f'{{bar:{self._empty_char}<{self._bar_length}}}',
			f'{{percentage:>{5 + self._percentage_precision}.{self._percentage_precision}%}}'
		])

	@property
	def height(self):
		'''
		The number of lines used by the progress bar.

		Returns
		-------
		height : int
			The number of lines used by the progress bar.
		'''

		return 1

	@property
	def width(self):
		'''
		The width of the progress bar.

		Returns
		-------
		width : int
			The complete width (counter + bar + percentage lengths).
		'''

		return len(self._pattern.format(counter = 0, bar = '', percentage = 0))

	@property
	def counter(self):
		'''
		The current value of the counter.

		Returns
		-------
		counter : int
			Counter value.
		'''

		return self._counter

	@UIDisplayedItem.renderer
	def render(self):
		'''
		Print the progress bar.
		'''

		percentage = self._counter / self._total
		n_full_chars = round(percentage * self._bar_length)

		print(self._pattern.format(counter = self._counter, bar = self._full_char * n_full_chars, percentage = percentage), end = '\r')

	@counter.setter
	def counter(self, n):
		'''
		Set the value of the counter.
		'''

		self.clear()
		self._counter = n
		self.render()

	def update(self, delta = 1):
		'''
		Set the value of the counter by adding an increment.

		Parameters
		----------
		delta : int
			Increment to add.
		'''

		self.counter += delta



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import json
import os
import shutil
import stat
import time

from .errors import *
from ..folder import Folder
from ..generator import Generator
from ..manager import Manager
from ..manager.errors import *
from ..remote import RemoteFolder
from ..remote.errors import *
from ..simulation import Simulation
from ..simulation.errors import *
from ..utils import Events, string, jsonfiles

class Maker():
	'''
	Assemble all components to extract simulations and automatically create them if they don't exist.

	Parameters
	----------
	simulations_folder : Folder|str
		The simulations folder. Must contain a settings file.

	config_name : str
		Name of the config to use. Indicate `None` to use the default configuration.

	override_options : dict
		Options to override.
	'''

	def __init__(self, simulations_folder, config_name = None, *, override_options = {}):
		self._simulations_folder = simulations_folder if type(simulations_folder) is Folder else Folder(simulations_folder)
		self._config_name = config_name

		self._manager_instance = None
		self._generator_instance = None
		self._remote_folder_instance = None

		self._loadOptions(override_options)

		self._simulations_to_extract = []
		self._unknown_simulations = []
		self._job_log_file = None

		self._remote_scripts_dir = None

		self._corruptions_counter = 0
		self._failures_counter = 0

		self._paused = False
		self._state_attrs = ['simulations_to_extract', 'corruptions_counter', 'failures_counter', 'unknown_simulations', 'remote_scripts_dir']

		self.events = Events([
			'close-start', 'close-end',
			'remote-open-start', 'remote-open-end',
			'delete-scripts',
			'paused', 'resume',
			'run-start', 'run-end',
			'extract-start', 'extract-end', 'extract-progress',
			'generate-start', 'generate-end',
			'wait-start', 'wait-progress', 'wait-end',
			'download-start', 'download-progress', 'download-end',
			'addition-start', 'addition-progress', 'addition-end'
		])

	def __enter__(self):
		'''
		Context manager to call `close()` at the end.
		'''

		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	@property
	def folder(self):
		'''
		Return the `Folder` instance.

		Returns
		-------
		folder : Folder
			The instance used by the maker.
		'''

		return self._simulations_folder

	@property
	def manager(self):
		'''
		Returns the instance of Manager used in the Maker.

		Returns
		-------
		manager : Manager
			Current instance, or a new one if `None`.
		'''

		if not(self._manager_instance):
			self._manager_instance = Manager(self._simulations_folder, readonly = self._options['generate_only'])

		return self._manager_instance

	@property
	def generator(self):
		'''
		Returns the instance of Generator used in the Maker.

		Returns
		-------
		generator : Generator
			Current instance, or a new one if `None`.
		'''

		if not(self._generator_instance):
			self._generator_instance = Generator(self._simulations_folder)

		return self._generator_instance

	@property
	def _remote_folder(self):
		'''
		Returns the instance of RemoteFolder used in the Maker.

		Returns
		-------
		remote_folder : RemoteFolder
			Current instance, or a new one if `None`.
		'''

		if not(self._remote_folder_instance):
			self._remote_folder_instance = RemoteFolder(self.folder.config('folder', self._config_name))

			self.events.trigger('remote-open-start')
			self._remote_folder_instance.open()
			self.events.trigger('remote-open-end')

		return self._remote_folder_instance

	def close(self):
		'''
		Clear all instances of the modules.
		'''

		self.events.trigger('close-start')

		self._generator_instance = None

		try:
			self._manager_instance.close()

		except AttributeError:
			pass

		try:
			self._remote_folder_instance.close()

		except AttributeError:
			pass

		self._remote_folder_instance = None

		self.events.trigger('close-end')

	def _loadOptions(self, override = {}):
		'''
		Load the options of the Maker, stored in the config folder.

		Parameters
		----------
		override : dict
			Options to impose the value of, despite the values in the config folder.
		'''

		self._options = {
			'settings_file': 'settings.json',
			'max_corrupted': -1,
			'max_failures': 0,
			'generate_only': False
		}

		try:
			self._options.update(self.folder.config('maker', self._config_name))

		except TypeError:
			pass

		self._options.update(override)

	def _setSimulations(self, simulations):
		'''
		Set the list of simulations to extract.

		Parameters
		----------
		simulations : list
			The list of simulations to extract.
		'''

		self._simulations_to_extract = [
			Simulation.ensureType(simulation, self._simulations_folder).copy()
			for simulation in simulations
		]

	@property
	def paused(self):
		'''
		Getter for the paused state.

		Returns
		-------
		paused : bool
			`True` if the Maker has been paused, `False` otherwise.
		'''

		return self._paused

	def pause(self):
		'''
		Pause the Maker.

		Raises
		------
		MakerPausedError
			The Maker is already in paused state.
		'''

		if self._paused:
			raise MakerPausedError()

		self._paused = True
		self.events.trigger('paused')

	def resume(self):
		'''
		Resume after a pause.

		Returns
		-------
		unknown_simulations : list
			List of simulations that failed to be generated. `None` if the Maker has been paused.

		Raises
		------
		MakerNotPausedError
			The Maker is not in paused state.
		'''

		if not(self._paused):
			raise MakerNotPausedError()

		self._paused = False
		self.events.trigger('resume')
		return self.run(self._simulations_to_extract)

	def saveState(self, filename):
		'''
		Save the current state of the Maker when it is paused.

		Parameters
		----------
		filename : str
			Name of the file to use to write the state.

		Raises
		------
		MakerNotPausedError
			The Maker is not in paused state.
		'''

		if not(self._paused):
			raise MakerNotPausedError()

		state = {attr: getattr(self, f'_{attr}') for attr in self._state_attrs}

		jsonfiles.write(state, filename)

	def loadState(self, filename):
		'''
		Load a state.

		Parameters
		----------
		filename : str
			Name of the file to use to read the state.

		Raises
		------
		MakerNotPausedError
			The Maker is not in paused state.

		MakerStateWrongFormatError
			At least one key is missing in the stored state.
		'''

		if not(self._paused):
			raise MakerNotPausedError()

		state = jsonfiles.read(filename)

		try:
			for attr in self._state_attrs:
				setattr(self, f'_{attr}', state[attr])

		except KeyError:
			raise MakerStateWrongFormatError()

	def run(self, simulations, *, corruptions_counter = 0, failures_counter = 0):
		'''
		Main loop, run until all simulations are extracted or some jobs failed.

		Parameters
		----------
		simulations : list
			List of simulations to extract/generate.

		corruptions_counter : int
			Initial value of the corruptions counter.

		failures_counter : int
		 	Initial value of the failures counter.

		Returns
		-------
		unknown_simulations : list
			List of simulations that failed to be generated. `None` if the Maker has been paused.
		'''

		self.events.trigger('run-start')

		self._setSimulations(simulations)

		self._corruptions_counter = corruptions_counter
		self._failures_counter = failures_counter

		while self._runLoop():
			pass

		if self.paused:
			return None

		self.events.trigger('run-end', self._unknown_simulations)

		return self._unknown_simulations

	def _runLoop(self):
		'''
		One loop of the `run()` method.

		Returns
		-------
		continue : bool
			`True` to continue the loop, `False` to break it.
		'''

		if self._job_log_file is None:
			self.extractSimulations()

			if not(self._unknown_simulations):
				return False

			if (self._options['max_corrupted'] >= 0 and self._corruptions_counter > self._options['max_corrupted']) or (self._options['max_failures'] >= 0 and self._failures_counter > self._options['max_failures']):
				return False

			self.generateSimulations()

		try:
			if not(self.waitForJob()):
				self._failures_counter += 1

		except KeyboardInterrupt:
			self.pause()
			return False

		if not(self.downloadSimulations()):
			self._corruptions_counter += 1

		self.events.trigger('delete-scripts')
		self._remote_folder.deleteRemote([self._remote_scripts_dir])

		return True

	def extractSimulations(self):
		'''
		Try to extract the simulations.
		'''

		self.events.trigger('extract-start', self._simulations_to_extract)

		self._unknown_simulations = self.manager.batchExtract(self._simulations_to_extract, settings_file = self._options['settings_file'], callback = lambda : self.events.trigger('extract-progress'))

		if self._options['generate_only']:
			self._unknown_simulations = list(filter(lambda simulation: not(os.path.isdir(simulation['folder'])), self._unknown_simulations))

		self.events.trigger('extract-end')

	def generateSimulations(self):
		'''
		Generate the scripts to generate the unknown simulations, and run them.

		Returns
		-------
		jobs_ids : list
			IDs of the jobs to wait.
		'''

		self.events.trigger('generate-start')

		scripts_dir = self._simulations_folder.tempdir()
		self._remote_scripts_dir = self._remote_folder.send(scripts_dir)

		self._simulations_to_generate = [simulation.copy() for simulation in self._unknown_simulations]

		self._simulations_remote_basedir = f'simulations_{hex(int(time.time() * 1E7))[2:]}'
		for k, simulation in enumerate(self._simulations_to_generate):
			simulation['folder'] = os.path.join(self._simulations_remote_basedir, str(k))

		self.generator.add(self._simulations_to_generate)
		script_to_launch = self.generator.generate(scripts_dir, self._config_name, empty_dest = True, basedir = self._remote_scripts_dir)

		self._job_log_file = self.generator.variables['LOG_FILENAME']

		self._remote_folder.send(scripts_dir, delete = True, replace = True)
		self._remote_folder.startServer(self.generator.variables['COMMAND_LINES_FILENAME'], self.generator.variables['JOB_DIRECTORY'], self._job_log_file)
		self._remote_folder.execute(script_to_launch)

		self.generator.clear()

		self.events.trigger('generate-end')

	def waitForJob(self):
		'''
		Wait for the job to finish.

		Returns
		-------
		success : bool
			`True` if the job has finished normally, `False` if there was at least one failure.
		'''

		n_total = len(self._simulations_to_generate)
		self.events.trigger('wait-start', n_total)

		n_finished = 0

		while True:
			try:
				log = json.loads(self._remote_folder.getFileContents(self._job_log_file))

			except FileNotFoundError:
				log = []

			if len(log) != n_finished:
				n_finished = len(log)
				self.events.trigger('wait-progress', n_finished)

				if n_finished == n_total:
					break

			time.sleep(0.5)

		self._job_log_file = None

		self.events.trigger('wait-end')

		return True

	def downloadSimulations(self):
		'''
		Download the generated simulations and add them to the manager.

		Returns
		-------
		success : bool
			`True` if all simulations has successfully been downloaded and added, `False` if there has been at least one issue.
		'''

		self.events.trigger('download-start', self._unknown_simulations)

		success = True

		for simulation, simulation_dest in zip(self._simulations_to_generate, self._unknown_simulations):
			tmpdir = self._simulations_folder.tempdir()
			try:
				self._remote_folder.receive(simulation['folder'], tmpdir, delete = True)

			except RemotePathNotFoundError:
				pass

			simulation['folder'] = tmpdir

			if self._options['generate_only']:
				if self._simulations_folder.checkIntegrity(simulation):
					destination_path = os.path.dirname(os.path.normpath(simulation_dest['folder']))
					if destination_path and not(os.path.isdir(destination_path)):
						os.makedirs(destination_path)

					os.rename(simulation['folder'], simulation_dest['folder'])

					if self._options['settings_file']:
						simulation_dest.writeSettingsFile(self._options['settings_file'])

				else:
					shutil.rmtree(simulation['folder'])
					success = False

			else:
				try:
					self.manager.add(simulation)

				except (SimulationFolderNotFoundError, SimulationIntegrityCheckFailedError):
					success = False

			self.events.trigger('download-progress')

		try:
			self._remote_folder.deleteRemote([self._simulations_remote_basedir])
		except FileNotFoundError:
			pass

		del self._simulations_remote_basedir
		del self._simulations_to_generate

		self.events.trigger('download-end')

		return success



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib.util
import pathlib
import sys
import uuid

from ..folder import Folder

def findFolder():
	'''
	Find the Hateno-compatible folder we're currently in by testing the existence of the `.hateno` subfolder.
	First search in the current folder, then in the parent, and then the parent of the parent, etc.

	Returns
	-------
	folder : Folder
		Folder instance corresponding to the found path. `None` if nothing has been found.
	'''

	folder = pathlib.Path('.').resolve()

	try:
		return Folder(folder)

	except FileNotFoundError:
		for parent in folder.parents:
			try:
				return Folder(parent)

			except FileNotFoundError:
				pass

	return None

def loadModuleFromFile(filename):
	'''
	Load a source file to use as a module.

	Parameters
	----------
	filename : str
		Path to the file to load.

	Returns
	-------
	module : module
		Loaded module.
	'''

	module_name = uuid.uuid4().hex
	spec = importlib.util.spec_from_file_location(module_name, filename)
	module = importlib.util.module_from_spec(spec)
	sys.modules[module_name] = module
	spec.loader.exec_module(module)

	return module



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class JobError(Exception):
	'''
	Base class for exceptions occurring in a job (server or client).
	'''

	pass

class JobDirAlreadyExistsError(JobError):
	'''
	Exception raised when we try to use a directory that already exists.

	Parameters
	----------
	job_dir : str
		The path to the directory.
	'''

	def __init__(self, job_dir):
		self.job_dir = job_dir



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re

from .errors import *
from .basesetting import SimulationBaseSetting

class SimulationSetting(SimulationBaseSetting):
	'''
	Represent a simulation setting.

	Parameters
	----------
	simulation : Simulation
		Simulation that uses this setting.

	set_name : str
		Name of the set this setting is part of.

	setting_name : str
		Name of this setting.

	Raises
	------
	SettingsSetNotFoundError
		The settings set has not been found.

	SettingNotFoundError
		The setting has not been found in the set.
	'''

	def __init__(self, simulation, set_name, setting_name):
		try:
			self._settings_set_dict = [
				settings_set
				for settings_set in simulation.folder.settings['settings']
				if settings_set['set'] == set_name
			][0]

		except IndexError:
			raise SettingsSetNotFoundError(set_name)

		try:
			self._setting_dict = [
				setting
				for setting in self._settings_set_dict['settings']
				if setting['name'] == setting_name
			][0]

		except IndexError:
			raise SettingNotFoundError(set_name, setting_name)

		super().__init__(simulation, setting_name, self._setting_dict['default'])

		self._set_name = set_name
		self._exclude_for_db = 'exclude' in self._setting_dict and self._setting_dict['exclude']
		self._pattern = self._setting_dict['pattern'] if 'pattern' in self._setting_dict else self._simulation.folder.settings['setting_pattern']

		self._use_only_if = 'only_if' in self._setting_dict
		if self._use_only_if:
			self._only_if_value = self._setting_dict['only_if']

		self._fixers_dict = None
		self._namers_dict = None

	def __str__(self):
		'''
		Return a string representing the setting and its value, according to the pattern.

		Returns
		-------
		setting : str
			The representation of the setting.
		'''

		value = self.value

		if type(value) is str and (not(value) or re.search(r'\s', value) is not None):
			value = repr(value)

		elif type(value) is list:
			value = ' '.join(map(str, value))

		return self._pattern.format(name = self.display_name, value = value)

	def as_dict(self):
		'''
		Dictionary representation of the setting.

		Returns
		-------
		setting : dict
			A dictionary listing some properties of the setting.
		'''

		return {
			'name': self.name,
			'value': self.value,
			'local_index': self._local_index,
			'local_total': self._simulation.getSettingCount(self.name, self._set_name),
			'global_index': self._global_index,
			'global_total': self._simulation.getSettingCount(self.name)
		}

	def setIndexes(self, global_index, local_index):
		'''
		Define the global and local indexes of this setting.

		Parameters
		----------
		global_index : int
			Global index of this setting.

		local_index : int
			Local index of this setting.
		'''

		self._global_index = global_index
		self._local_index = local_index

	def _setModifier(self, modifier):
		'''
		Set a "modifier" property (fixers or namers).

		Parameters
		----------
		modifier : str
			Modifier to define.

		Returns
		-------
		modifier_functions : dict
			The functions to call, in the right order.
		'''

		modifier_functions = {'before': [], 'after': []}

		keys_to_search = {
			'before': [
				(self._setting_dict, ''),
				(self._settings_set_dict, ''),
				(self._setting_dict, '_before'),
				(self._settings_set_dict, '_before'),
				(self._setting_dict, '_between_before')
			],
			'after': [
				(self._setting_dict, '_between_after'),
				(self._settings_set_dict, '_after'),
				(self._setting_dict, '_after')
			]
		}

		for when, keys_path in keys_to_search.items():
			for dict_to_search, key_to_search in keys_path:
				try:
					modifier_functions[when] += dict_to_search[f'{modifier}{key_to_search}']

				except KeyError:
					pass

		return modifier_functions

	@property
	def _fixers(self):
		'''
		Get the fixers to apply to the value.

		Returns
		-------
		fixers : dict
			The list of fixers, in the right order.
		'''

		if self._fixers_dict is None:
			self._fixers_dict = self._setModifier('fixers')

		return self._fixers_dict

	@property
	def _namers(self):
		'''
		Get the namers to apply to the name.

		Returns
		-------
		namers : dict
			The list of namers, in the right order.
		'''

		if self._namers_dict is None:
			self._namers_dict = self._setModifier('namers')

		return self._namers_dict

	@property
	def display_name(self):
		'''
		Get the name of the setting to use inside the simulation.

		Returns
		-------
		name : str
			Name to use.
		'''

		return self._simulation.folder.applyNamers(self.as_dict(), **self._namers)

	@property
	def exclude(self):
		'''
		Return `True` if the setting should be excluded to define a simulation.

		Returns
		-------
		exclude : bool
			Exclusion parameter.
		'''

		return self._exclude_for_db

	def shouldBeDisplayed(self):
		'''
		Return `True` if the setting can be displayed/used for the execution.
		This check is based on the `only_if` parameter. There are several cases for the test string.

		1. It is a full test (e.g. `a < b`): we execute it as it is.
		2. It begins by a conditional operator (<, <=, >, >=, ==, !=, in): we test against the setting value.
		3. It does not contain any conditional operator: we test the equality against the setting value.

		Returns
		-------
		should_be_displayed : bool
			The result of the check.
		'''

		if not(self._use_only_if):
			return True

		if type(self._only_if_value) is str:
			first_operator_match = re.search(r'([<>]=?|[!=]=|in)', self._only_if_value.strip())

			if first_operator_match:
				if first_operator_match.start() > 0:
					return self._simulation.parseString(f'(({self._only_if_value}))')

				return self._simulation.parseString(f'(({repr(self.value)} {self._only_if_value}))')

			return self.value == self._simulation.parseString(self._only_if_value)

		return self.value == self._only_if_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import subprocess
import time

import watchdog.events

from .filewatcher import FileWatcher
from ..utils import Events, jsonfiles, string

class JobClient():
	'''
	Client part of a job, executing the command lines sent by the server.

	Parameters
	----------
	job_dir : str
		Path to the job directory where the messaging files are stored.
	'''

	def __init__(self, job_dir):
		self._job_dir = job_dir

		self.events = Events([
			'received', 'sent'
		])

	def __enter__(self):
		'''
		Context manager to automatically call `stop()`.
		'''

		self.start()
		return self

	def __exit__(self, *args, **kwargs):
		'''
		Call `stop()` when exiting a context manager.
		'''

		self.stop()

	def start(self):
		'''
		Create the messaging file for this client.
		'''

		self._client_id = string.uniqueID()
		self._client_filename = os.path.join(self._job_dir, f'{self._client_id}.json')

		self._event_handler = FileEventHandler(self._processResponse)

		with open(self._client_filename, 'w') as f:
			f.write('')

	def stop(self):
		'''
		Delete the messaging file of the client.
		'''

		try:
			os.unlink(self._client_filename)

		except FileNotFoundError:
			pass

	def _processResponse(self, response):
		'''
		Handle the response from the server.
		If the server sent a command line, execute it and send the log.
		If the "command line" is `None`, stop everything.

		Parameters
		----------
		response : dict
			Message sent by the server.
		'''

		self._wait = False

		self.events.trigger('received', response['command_line'])

		if response['command_line'] is not None:
			p = subprocess.run(response['command_line'], shell = True, capture_output = True, encoding = 'utf-8')

			log = {
				'log': {
					'exec': response['command_line'],
					'stdout': p.stdout,
					'stderr': p.stderr,
					'success': p.returncode == 0
				},
				'state': 'ready'
			}

			self._sendAndWait(log)

	def _sendAndWait(self, req):
		'''
		Send a message to the server and wait for the response.

		Parameters
		----------
		req : dict
			Request to send.
		'''

		try:
			with FileWatcher(self._event_handler, self._client_filename):
				jsonfiles.write(req, self._client_filename)
				self.events.trigger('sent', req)

				self._wait = True
				while self._wait:
					time.sleep(0.1)

		except FileNotFoundError:
			pass

	def run(self):
		'''
		Run "loop". Tell the server the client is ready and wait for the first command line.
		'''

		self._sendAndWait({'state': 'ready'})

class FileEventHandler(watchdog.events.FileSystemEventHandler):
	'''
	Handle the events from the files modifications.

	Parameters
	----------
	response : function
		Function to call to handle a response from the server.
	'''

	def __init__(self, response):
		self._response = response

		self._prev_message = None

	def on_modified(self, evt):
		'''
		The client file has been modified.
		Read the content of the file and test whether it is a message from the server.

		Parameters
		----------
		evt : watchdog.events.FileModifiedEvent
			Object given by Watchdog containing infos about the modified file.
		'''

		try:
			content = jsonfiles.read(evt.src_path)

		except:
			pass

		else:
			if content == self._prev_message:
				return

			self._prev_message = content

			if 'command_line' in content:
				self._response(content)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class MakerError(Exception):
	'''
	Base class for exceptions occurring in the Maker.
	'''

	pass

class MakerPausedError(MakerError):
	'''
	Exception raised when the Maker is paused while it should not be.
	'''

	pass

class MakerNotPausedError(MakerError):
	'''
	Exception raised when the Maker is not paused while it should be.
	'''

	pass

class MakerStateWrongFormatError(MakerError):
	'''
	Exception raised when we try to read a Maker state that is in the wrong format.
	'''

	pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .item import UIDisplayedItem

class UITextLine(UIDisplayedItem):
	'''
	Represent a text line displayed in the UI.

	Parameters
	----------
	ui : UI
		The UI object this text line belongs to.

	text : str
		The text to display.
	'''

	def __init__(self, ui, text):
		super().__init__(ui)

		self._text = text

	@property
	def height(self):
		'''
		The number of lines used by the text line.
		Currently, always one. Multilines are not supported yet.

		Returns
		-------
		height : int
			The number of lines used by the text.
		'''

		return 1

	@property
	def width(self):
		'''
		The width of the text line, i.e. the length of the text.

		Returns
		-------
		width : int
			The length of the text.
		'''

		return len(self.text)

	@property
	def text(self):
		'''
		Getter for the displayed text.

		Returns
		-------
		text : str
			Displayed text.
		'''

		return self._text

	@UIDisplayedItem.renderer
	def render(self):
		'''
		Print the text.
		'''

		print(self._text, end = '\r')

	@text.setter
	def text(self, new_text):
		'''
		Change the displayed text.

		Parameters
		----------
		new_text : str
			New text to display.
		'''

		self.clear()
		self._text = new_text
		self.render()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import paramiko
import stat

class SFTP(paramiko.SFTPClient):
	'''
	Overwrite some methods to add some options.
	'''

	def chdir(self, wd):
		'''
		Change the current working directory.
		If it does not exist yet, create it.

		Parameters
		----------
		wd : str
			Working directory to use.
		'''

		try:
			s = self.stat(wd)

		except FileNotFoundError:
			self._makedirs(wd)

		finally:
			super().chdir(wd)

	def _copyLocalChmod(self, local_path, remote_path):
		'''
		Change the chmod of a remote file/folder to reflect a local one.

		Parameters
		----------
		local_path : str
			Name of the file/folder to use to copy the chmod.

		remote_path : str
			Remote path to alter.
		'''

		self.chmod(remote_path, os.stat(local_path).st_mode & 0o777)

	def _copyRemoteChmod(self, remote_path, local_path):
		'''
		Change the chmod of a local file/folder to reflect a remote one.

		Parameters
		----------
		remote_path : str
			Name of the file/folder to use to copy the chmod.

		local_path : str
			Local path to alter.
		'''

		os.chmod(local_path, self.stat(remote_path).st_mode & 0o777)

	def _makedirs(self, directory):
		'''
		Recursively create a directory.

		Parameters
		----------
		directory : str
			Path to create.
		'''

		try:
			self.mkdir(directory)

		except FileNotFoundError:
			self._makedirs(os.path.dirname(os.path.normpath(directory)))
			self.mkdir(directory)

	def put(self, local_path, remote_path, replace = False, delete = False):
		'''
		Send a file or a folder.

		Parameters
		----------
		local_path : str
			Path to the file/folder to send.

		remote_path : str
			Path of the remote file/folder to create.

		replace : bool
			If `False`, send a file only if the source is more recent. Otherwise always send it.

		delete : bool
			If `True`, delete the local file/folder once sent.
		'''

		if os.path.isfile(local_path):
			if not(replace):
				try:
					if os.stat(local_path).st_mtime <= self.stat(remote_path).st_mtime:
						return None

				except FileNotFoundError:
					pass

			try:
				super().put(local_path, remote_path)

			except FileNotFoundError:
				self._makedirs(os.path.dirname(remote_path))
				super().put(local_path, remote_path)

			self._copyLocalChmod(local_path, remote_path)

			if delete:
				os.unlink(local_path)

		else:
			for entry in os.listdir(local_path):
				self.put(os.path.join(local_path, entry), os.path.join(remote_path, entry), replace, delete)

			if delete:
				os.rmdir(local_path)

	def get(self, remote_path, local_path, delete = False):
		'''
		Download a file or a folder.

		Parameters
		----------
		remote_path : str
			Path to the file/folder to download.

		local_path : str
			Path of file/folder to create.

		delete : bool
			If `True`, delete the remote file/folder once downloaded.
		'''

		if stat.S_ISDIR(self.stat(remote_path).st_mode):
			for entry in self.listdir(remote_path):
				self.get(os.path.join(remote_path, entry), os.path.join(local_path, entry), delete)

			if delete:
				self.rmdir(remote_path)

		else:
			try:
				super().get(remote_path, local_path)

			except FileNotFoundError:
				os.makedirs(os.path.dirname(local_path))
				super().get(remote_path, local_path)

			self._copyRemoteChmod(remote_path, local_path)

			if delete:
				self.remove(remote_path)

	def remove(self, path):
		'''
		Remove a file or a folder.

		Parameters
		----------
		path : str
			Path of the file/folder to remove.
		'''

		if stat.S_ISDIR(self.stat(path).st_mode):
			for entry in self.listdir(path):
				self.remove(os.path.join(path, entry))

			self.rmdir(path)

		else:
			super().remove(path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
import operator
import os
import re
import shutil

from ..folder import Folder
from ..maker import Maker
from ..simulation import Simulation
from ..utils import Events, string, jsonfiles

class Mapper():
	'''
	Use the Maker to generate simulations and associate them to numeric values.

	Parameters
	----------
	simulations_folder : Folder|str
		The simulations folder. Either a `Folder` instance or a path to a folder with a configuration file.

	config_name : str
		Name of the config to use with the Maker.

	generate_only : bool
		`True` to not add the simulations to the manager.
	'''

	def __init__(self, simulations_folder, config_name = None, *, generate_only = True):
		self._simulations_folder = simulations_folder if type(simulations_folder) is Folder else Folder(simulations_folder)
		self._config_name = config_name

		self._generate_only = generate_only
		self._maker_instance = None

		self._default_simulation = None
		self._tree = None
		self._output = None

		self._simulations_dir = None

		self._index_regex_compiled = None

		self.events = Events([
			'read-start', 'read-end',
			'map-start', 'map-end',
			'node-start', 'node-progress', 'node-end',
			'generate-start', 'generate-end',
			'evaluation-start', 'evaluation-end',
			'explorer-find-start', 'explorer-find-end',
			'explorer-searches-start', 'explorer-searches-end',
			'explorer-search-start', 'explorer-search-end',
			'explorer-search-iteration-start', 'explorer-search-iteration-end'
		])

	def __enter__(self):
		'''
		Context manager to call `close()` at the end.
		'''

		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	@property
	def maker(self):
		'''
		Return the Maker instance.

		Returns
		-------
		maker : Maker
			Instance currently in used.
		'''

		if self._maker_instance is None:
			self._maker_instance = Maker(self._simulations_folder, self._config_name, override_options = {'generate_only': self._generate_only})

		return self._maker_instance

	def close(self):
		'''
		Properly exit the Maker instance.
		'''

		try:
			self._maker_instance.close()

		except AttributeError:
			pass

		else:
			self._maker_instance = None

	@property
	def tree(self):
		'''
		Get the current tree.

		Returns
		-------
		tree : dict
			The current tree to read.
		'''

		return self._tree

	@property
	def default_simulation(self):
		'''
		Get the current default simulation's settings.

		Returns
		-------
		default_simulation : dict
			The settings of the current default simulation.
		'''

		return self._default_simulation.raw_values_settings

	@property
	def output(self):
		'''
		Get the current output.

		Returns
		-------
		output : dict
			The output of the current tree.
		'''

		return self._output

	@property
	def tree_by_depths(self):
		'''
		Get the current tree, sorted by depths.

		Returns
		-------
		tree : dict
			The current tree to read.
		'''

		if self._tree is None:
			return None

		depth = 0
		tree = {depth: self._tree}

		current_depth = self._tree
		while 'foreach' in current_depth:
			depth += 1
			current_depth = current_depth['foreach']
			tree[depth] = current_depth

		return tree

	@property
	def _index_regex(self):
		'''
		Regex to detect an index in a test.

		Returns
		-------
		regex : re.Pattern
			The index regex.
		'''

		if self._index_regex_compiled is None:
			self._index_regex_compiled = re.compile(r'\[(-?[0-9]+)\]')

		return self._index_regex_compiled

	def _buildValues(self, vdesc):
		'''
		Build the list of values a node will use.
		For values description, we don't calculate the interpolated values now. The Simulation parser will, so we can use settings tags.

		Parameters
		----------
		vdesc : list|dict
			Explicit list, or description. A description is a dictionary with the following keys:
				* `from`: the first value(s),
				* `to`: the last value(s),
				* `n`: the number of values.

		Returns
		-------
		values : list
			The complete list of values for the node.
		'''

		if type(vdesc) is list:
			return [
				v if type(v) is list else [v]
				for v in vdesc
			]

		if type(vdesc['from']) is list:
			return [
				[
					f'(({a} + {k} * ({b} - ({a})) / {vdesc["n"] - 1}))'
					for a, b in zip(vdesc['from'], vdesc['to'])
				]
				for k in range(0, vdesc['n'])
			]

		return [
			[f'(({vdesc["from"]} + {k} * ({vdesc["to"]} - ({vdesc["from"]})) / {vdesc["n"] - 1}))']
			for k in range(0, vdesc['n'])
		]

	def _readTreeNode(self, node):
		'''
		Normalize the way a node is defined, i.e. always use `settings` with a list, and build the values.
		Detect the depth at which we will generate the simulations (the shallowest one before finding a stop).

		Parameters
		----------
		node : dict
			Description of the node.

		Returns
		-------
		normalized : dict
			The normalized node.
		'''

		settings = node.get('settings') or node.get('setting')
		if type(settings) is not list:
			settings = [settings]

		normalized = {
			'settings': [{'set': '', 'set_index': 0, 'name': '', **coords} for coords in settings],
			'values': self._buildValues(node['values'])
		}

		depth = len(self._nodes)
		self._nodes.append(normalized)

		self._simulations_settings.append(normalized['settings'])
		self._simulations_settings_values.append(normalized['values'])

		if 'foreach' in node:
			normalized['foreach'] = self._readTreeNode(node['foreach'])

		try:
			normalized['evaluation'] = node['evaluation']

		except KeyError:
			pass

		try:
			normalized['test'] = node['test']

		except KeyError:
			pass

		else:
			try:
				normalized['test'] = node['test_target']

			except KeyError:
				pass

			if isinstance(normalized['test'], (float, int)):
				normalized['test_target'] = normalized['test']
				normalized['test'] = f'([-2] - {normalized["test"]}) * ([-1] - {normalized["test"]}) <= 0'

			else:
				first_operator_match = re.search(r'([<>]=?|[!=]=|in)', normalized['test'].strip())

				if first_operator_match and first_operator_match.start() == 0:
					normalized['test'] = '[-1] ' + normalized['test']

			normalized['test_requested_indices'] = set(map(int, self._index_regex.findall(normalized['test'])))

		try:
			normalized['stop'] = node['stop']

		except KeyError:
			normalized['stop'] = False

		if normalized['stop']:
			self._generating_depth = max(self._generating_depth, depth + 1)

		return normalized

	def _readTree(self, tree):
		'''
		Define the default simulation and normalize the way the tree is defined.

		Parameters
		----------
		tree : dict
			Description of the tree to explore.
		'''

		self.events.trigger('read-start')

		self._default_simulation = Simulation(self._simulations_folder, {'settings': tree.get('default') or {}})
		self._simulations_settings = []
		self._simulations_settings_values = []
		self._generating_depth = 0
		self._nodes = []
		self._tree = self._readTreeNode(tree['tree'])

		self.events.trigger('read-end')

	def _generateSimulations(self, prev_values):
		'''
		Generate a set of simulations, corresponding to all children of a node.

		Parameters
		----------
		prev_values : tuple
			Indices of the values of the settings of the parent nodes.
		'''

		self.events.trigger('generate-start')

		if self._simulations_dir is None:
			self._simulations_dir = self._simulations_folder.tempdir()
			self._simulations = {}

		simulations_dir = os.path.join(self._simulations_dir, str(len(os.listdir(self._simulations_dir))))

		default_simulation = self._default_simulation.copy()
		for settings, values, k in zip(self._simulations_settings, self._simulations_settings_values, prev_values):
			for setting, value in zip(settings, values[k]):
				default_simulation.getSetting(setting).value = value

		simulations = []
		remaining_settings = self._simulations_settings[len(prev_values):]

		if remaining_settings:
			remaining_values = self._simulations_settings_values[len(prev_values):]
			remaining_values_indices = map(lambda l: range(len(l)), remaining_values)

			indices_coefs = tuple(itertools.accumulate([1] + list(reversed(list(map(lambda l: len(l), remaining_values[1:]))))))

			for values, K in zip(itertools.product(*remaining_values), itertools.product(*remaining_values_indices)):
				k = sum(map(operator.mul, indices_coefs, K))

				simulation = default_simulation.copy()
				simulation['folder'] = os.path.join(simulations_dir, str(k))

				for setting, value in zip(sum(remaining_settings, []), sum(values, [])):
					simulation.getSetting(setting).value = value

				simulations.append(simulation)
				self._simulations[prev_values + K] = simulation

		else:
			default_simulation['folder'] = os.path.join(simulations_dir, '0')
			simulations = [default_simulation]
			self._simulations[prev_values] = default_simulation

		self.maker.run(simulations)

		self.events.trigger('generate-end')

	def _deleteSimulations(self):
		'''
		Delete the simulations directory.
		'''

		if self._simulations_dir is not None:
			shutil.rmtree(self._simulations_dir)

			self._simulations_dir = None
			self._simulations = None

	def _evaluate(self, node, simulations):
		'''
		Evaluate the simulations of a node, and store the output.

		Parameters
		----------
		node : dict
			The current node.

		Simulations : list
			The list of simulations corresponding to the node.

		Returns
		-------
		evaluation : mixed
			Result of the evaluation, or `None` if there is no evaluation found in the node.
		'''

		if 'evaluation' not in node:
			return None

		self.events.trigger('evaluation-start')

		arg_to_pass = simulations if 'foreach' in node else simulations[0]
		evaluation = self._simulations_folder.evaluations.call(node['evaluation'], arg_to_pass)

		self.events.trigger('evaluation-end')

		return evaluation

	def _test(self, node, evaluations):
		'''
		Apply a test defined in a node.

		Parameters
		----------
		node : dict
			The current node.

		evaluations : list
			List of evaluations to consider.

		Returns
		-------
		result : mixed
			Result of the test (bool) or `None` if no test has been evaluated.
		'''

		if 'test' not in node:
			return None

		test = node['test']

		# Retrieve the requested indices and first see if the ones which should be different are really different
		# To do that, we "fix" negative indexes so they should be equal to their positive counterparts
		# If they are still negative, it seems that we don't have enough elements in the list: it will be detected in the next part
		# It is important to add the length of the list and to not use the modulo, to be sure we consider different elements

		unique_requested_indices = set(map(lambda k: k if k >= 0 else k + len(evaluations), node['test_requested_indices']))

		if len(unique_requested_indices) < len(node['test_requested_indices']):
			return None

		try:
			test = self._index_regex.sub(lambda m: str(evaluations[int(m.group(1))]), test)

		except IndexError:
			return None

		else:
			return string.safeEval(test)

	def _endNode(self, node, simulations, evaluations, output):
		'''
		Actions to execute at the end of the mapping of a node.
		First, evaluate the generated simulation(s).
		Then, apply the test.

		Parameters
		----------
		node : dict
			The mapped node.

		simulations : list
			The list of simulations to evaluate.

		evaluations : list
			The current list of evaluations for this node.

		output : dict
			Output of the node.

		Returns
		-------
		stop : bool
			`True` if the latest test is `True` and the node's `stop` is `True`.
		'''

		evaluation = self._evaluate(node, simulations)

		if evaluation is None:
			return False

		evaluations.append(evaluation)
		output['evaluation'] = evaluation

		test = self._test(node, evaluations)

		if test is None:
			return False

		output['test'] = test

		return node['stop'] and test

	def _mapNode(self, output, prev_values = ()):
		'''
		Map a node by associating evaluations to the right simulations.
		Generate the simulations if we are located at the right depth.

		Parameters
		----------
		output : dict
			Dictionary where to store the output of this node.

		prev_values : tuple
			Indices of the values for the parameters of the parent nodes.

		Returns
		-------
		simulations : list
			The list of simulations corresponding to this node.
		'''

		depth = len(prev_values)

		try:
			node = self._nodes[depth]

		except IndexError:
			node = None

		self.events.trigger('node-start', depth, node)

		if depth == self._generating_depth:
			self._generateSimulations(prev_values)

		if node is None:
			return [self._simulations[prev_values]]

		output['settings'] = node['settings']
		output['map'] = []

		simulations = []
		evaluations = []

		for k in range(len(node['values'])):
			sub_output = {'values': [], 'output': {}}

			sub_simulations = self._mapNode(sub_output['output'], prev_values + (k,))
			simulations += sub_simulations

			sub_output['values'] = [sub_simulations[0].getSetting(setting).value for setting in node['settings']]

			if not(sub_output['output']):
				del sub_output['output']

			stop = self._endNode(node, sub_simulations, evaluations, sub_output)
			output['map'].append(sub_output)

			self.events.trigger('node-progress', depth)

			if stop:
				break

		self.events.trigger('node-end', depth)

		return simulations

	def mapTree(self, tree):
		'''
		Map a tree.

		Parameters
		----------
		tree : dict
			Description of the tree to explore.

		Returns
		-------
		output : dict
			Output of the tree mapping.
		'''

		self._readTree(tree)

		self.events.trigger('map-start')

		self._output = {}
		self._mapNode(self._output)

		self._deleteSimulations()

		self.events.trigger('map-end')

		return self._output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class SimulationError(Exception):
	'''
	Base class for exceptions occurring in a Simulation.
	'''

	pass

class SettingNotFoundError(SimulationError):
	'''
	Exception raised when a setting has not been found in a given set.

	Parameters
	----------
	set_name : str
		Name of the set.

	setting_name : str
		Name of the setting
	'''

	def __init__(self, set_name, setting_name):
		self.set_name = set_name
		self.setting_name = setting_name


class SettingsSetNotFoundError(SimulationError):
	'''
	Exception raised when a settings set has not been found.

	Parameters
	----------
	set_name : str
		Name of the set.
	'''

	def __init__(self, set_name):
		self.set_name = set_name

class SettingTagNotRecognizedError(SimulationError):
	'''
	Exception raised when we try to analyse a setting tag but it fails.

	Parameters
	----------
	setting_tag : str
		The invalid tag.
	'''

	def __init__(self, setting_tag):
		self.setting_tag = setting_tag



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class UIError(Exception):
	'''
	Base class for exceptions occurring in the UI.
	'''

	pass

class UINonMovableLine(UIError):
	'''
	Exception raised when we try to move a line which can't be moved.

	Parameters
	----------
	pos : int
		Position of the line.
	'''

	def __init__(self, pos):
		self.pos = pos



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..ui.ui import UI
from ..utils import string

class MakerUI(UI):
	'''
	UI to show the different steps of the Maker.

	Parameters
	----------
	maker : Maker
		Instance of the Maker from which the event are triggered.
	'''

	def __init__(self, maker):
		super().__init__()

		self._maker = maker

		self._state_line = None
		self._main_progress_bar = None

		self._maker.events.addListener('close-start', self._closeStart)
		self._maker.events.addListener('close-end', self._closeEnd)
		self._maker.events.addListener('remote-open-start', self._remoteOpenStart)
		self._maker.events.addListener('remote-open-end', self._remoteOpenEnd)
		self._maker.events.addListener('delete-scripts', self._deleteScripts)
		self._maker.events.addListener('paused', self._paused)
		self._maker.events.addListener('resume', self._resume)
		self._maker.events.addListener('run-start', self._runStart)
		self._maker.events.addListener('run-end', self._runEnd)
		self._maker.events.addListener('extract-start', self._extractStart)
		self._maker.events.addListener('extract-progress', self._extractProgress)
		self._maker.events.addListener('extract-end', self._extractEnd)
		self._maker.events.addListener('generate-start', self._generateStart)
		self._maker.events.addListener('generate-end', self._generateEnd)
		self._maker.events.addListener('wait-start', self._waitStart)
		self._maker.events.addListener('wait-progress', self._waitProgress)
		self._maker.events.addListener('wait-end', self._waitEnd)
		self._maker.events.addListener('download-start', self._downloadStart)
		self._maker.events.addListener('download-progress', self._downloadProgress)
		self._maker.events.addListener('download-end', self._downloadEnd)

	def _updateState(self, state):
		'''
		Text line to display the current state of the Maker.

		Parameters
		----------
		state : str
			State to display.
		'''

		if self._state_line is None:
			self._state_line = self.addTextLine(state)

		else:
			self._state_line.text = state

	def _clearState(self):
		'''
		Remove the state line.
		'''

		if not(self._state_line is None):
			self.removeItem(self._state_line)
			self._state_line = None

	def _closeStart(self):
		'''
		Maker starts closing.
		'''

		pass

	def _closeEnd(self):
		'''
		Maker is closed.
		'''

		pass

	def _remoteOpenStart(self):
		'''
		Connection to the RemoteFolder is started.
		'''

		self._updateState('Connection…')

	def _remoteOpenEnd(self):
		'''
		Connected to the RemoteFolder.
		'''

		self._updateState('Connected')

	def _deleteScripts(self):
		'''
		Deletion of the scripts.
		'''

		self._updateState('Deleting the scripts…')

	def _paused(self):
		'''
		The Maker has been paused.
		'''

		# Erase the "^C" due to the keyboard interruption
		print('\r  ', end = '\r')

		self._updateState('Paused')

	def _resume(self):
		'''
		Resume after a pause.
		'''

		pass

	def _runStart(self):
		'''
		The run loop just started.
		'''

		self._updateState('Running the Maker…')

	def _runEnd(self, unknown_simulations):
		'''
		The run loop has ended.

		Parameters
		----------
		unknown_simulations : list
			List of simulations that still do not exist.
		'''

		if unknown_simulations:
			self._updateState(string.plural(len(unknown_simulations), 'simulation still does not exist', 'simulations still do not exist'))

		else:
			self._updateState('All simulations have successfully been extracted')

	def _extractStart(self, simulations):
		'''
		Start the extraction of the simulations.

		Parameters
		----------
		simulations : list
			List of the simulations that will be extracted.
		'''

		self._updateState('Extracting the simulations…')
		self._main_progress_bar = self.addProgressBar(len(simulations))

	def _extractProgress(self):
		'''
		A simulation has just been extracted.
		'''

		self._main_progress_bar.counter += 1

	def _extractEnd(self):
		'''
		All simulations have been extracted.
		'''

		self.removeItem(self._main_progress_bar)
		self._main_progress_bar = None
		self._updateState('Simulations extracted')

	def _generateStart(self):
		'''
		Start the generation of the scripts.
		'''

		self._updateState('Generating the scripts…')

	def _generateEnd(self):
		'''
		Scripts are generated.
		'''

		self._updateState('Scripts generated')

	def _waitStart(self, n_simulations):
		'''
		Start to wait for a job.

		Parameters
		----------
		n_simulations : int
			Number of simulations in the job.
		'''

		simulations = string.plural(n_simulations, 'simulation', 'simulations')
		self._updateState(f'Waiting for {simulations} to execute…')
		self._main_progress_bar = self.addProgressBar(n_simulations)

	def _waitProgress(self, n_executed):
		'''
		Update the number of executed simulations.

		Parameters
		----------
		n_executed : int
			Number of executed simulations.
		'''

		self._main_progress_bar.counter = n_executed

	def _waitEnd(self):
		'''
		The job has finished.
		'''

		self.removeItem(self._main_progress_bar)
		self._main_progress_bar = None

		self._updateState('Simulations executed')

	def _downloadStart(self, simulations):
		'''
		Start to download and add the simulations.

		Parameters
		----------
		simulations : list
			Simulations that will be downloaded.
		'''

		self._updateState('Downloading the simulations…')
		self._main_progress_bar = self.addProgressBar(len(simulations))

	def _downloadProgress(self):
		'''
		A simulation has just been downloaded and added.
		'''

		self._main_progress_bar.counter += 1

	def _downloadEnd(self):
		'''
		All simulations have been downloaded and added.
		'''

		self.removeItem(self._main_progress_bar)
		self._main_progress_bar = None
		self._updateState('Simulations downloaded')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..maker import MakerUI
from ..utils import string

class MapperUI(MakerUI):
	'''
	UI to show the different steps of the Mapper.

	Parameters
	----------
	mapper : Mapper
		Instance of the Mapper from which the events are triggered.
	'''

	def __init__(self, mapper):
		super().__init__(mapper.maker)

		self._mapper = mapper

		self._mapper_state_line = None

		self._nodes_lines = {}
		self._nodes_bars = {}

		self._explorer_searches = None
		self._explorer_searches_n_tests = None
		self._explorer_search = None

		self._mapper.events.addListener('read-start', self._readStart)
		self._mapper.events.addListener('read-end', self._readEnd)
		self._mapper.events.addListener('map-start', self._mapStart)
		self._mapper.events.addListener('map-end', self._mapEnd)
		self._mapper.events.addListener('node-start', self._nodeStart)
		self._mapper.events.addListener('node-progress', self._nodeProgress)
		self._mapper.events.addListener('node-end', self._nodeEnd)
		self._mapper.events.addListener('generate-start', self._generateStart)
		self._mapper.events.addListener('generate-end', self._generateEnd)
		self._mapper.events.addListener('evaluation-start', self._evaluationStart)
		self._mapper.events.addListener('evaluation-end', self._evaluationEnd)
		self._mapper.events.addListener('explorer-find-end', self._explorerFindStart)
		self._mapper.events.addListener('explorer-find-end', self._explorerFindEnd)
		self._mapper.events.addListener('explorer-search-start', self._explorerSearchStart)
		self._mapper.events.addListener('explorer-search-end', self._explorerSearchEnd)
		self._mapper.events.addListener('explorer-searches-start', self._explorerSearchesStart)
		self._mapper.events.addListener('explorer-searches-end', self._explorerSearchesEnd)
		self._mapper.events.addListener('explorer-search-iteration-start', self._explorerSearchIterationStart)
		self._mapper.events.addListener('explorer-search-iteration-end', self._explorerSearchIterationEnd)

		self._mapper.maker.events.addListener('run-end', self._clearMakerState)

	def _updateMapperState(self, state):
		'''
		Text line to display the current state of the Mapper.

		Parameters
		----------
		state : str
			State to display.
		'''

		if self._mapper_state_line is None:
			self._mapper_state_line = self.addTextLine(state, position = 0)

		else:
			self._mapper_state_line.text = state

	def _clearMakerState(self, *args, **kwargs):
		'''
		We don't need the Maker state anymore: we erase it.
		'''

		self._clearState()

	def clearState(self):
		'''
		Clear the Mapper state.
		'''

		if self._mapper_state_line is not None:
			self.removeItem(self._mapper_state_line)
			self._mapper_state_line = None

	def _readStart(self):
		'''
		The interpretation of a tree has started.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Reading a tree…')

	def _readEnd(self):
		'''
		The tree is read.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Tree read')

	def _mapStart(self):
		'''
		The mapping of a tree has been started.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Mapping a tree…')

	def _mapEnd(self):
		'''
		The mapping of a tree has ended.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Tree mapped')

	def _nodeStart(self, depth, node):
		'''
		The mapping of a node has been started.

		Parameters
		----------
		depth : int
			Depth of the node.

		node : dict
			Description of the node.
		'''

		if node is not None:
			self._nodes_lines[depth] = self.addTextLine(f'Depth {depth}…', position = 2*depth + 1)
			self._nodes_bars[depth] = self.addProgressBar(len(node['values']), position = 2*depth + 2)

	def _nodeProgress(self, depth):
		'''
		The mapping of a node just progressed.

		Parameters
		----------
		depth : int
			Depth of the node.
		'''

		if depth in self._nodes_bars:
			self._nodes_bars[depth].counter += 1

	def _nodeEnd(self, depth):
		'''
		The mapping of a node has ended.

		Parameters
		----------
		depth : int
			Depth of the node.
		'''

		if depth in self._nodes_lines:
			self.removeItem(self._nodes_bars[depth])
			del self._nodes_bars[depth]

			self.removeItem(self._nodes_lines[depth])
			del self._nodes_lines[depth]

	def _generateStart(self):
		'''
		A set of simulations will be generated.
		'''

		pass

	def _generateEnd(self):
		'''
		A set of simulations has been generated.
		'''

		pass

	def _evaluationStart(self):
		'''
		The evaluation of a simulation has been started.
		'''

		pass

	def _evaluationEnd(self):
		'''
		The evaluation of a simulation has ended.
		'''

		pass

	def _explorerFindStart(self):
		'''
		An Explorer starts to find the tests evaluated to `True`.
		'''

		self._updateMapperState('Finding the true tests…')

	def _explorerFindEnd(self):
		'''
		The search for the tests has ended.
		'''

		self._updateMapperState('Exploration ended')

	def _explorerSearchesStart(self, searches, n_tests):
		'''
		An Explorer starts to search optimal settings.

		Parameters
		----------
		searches : dict
			Description of all the searches.

		n_tests : int
			Number of planned searches.
		'''

		self._explorer_searches = searches
		self._explorer_searches_n_tests = n_tests

		self._updateMapperState(f'Searching for {string.plural(n_tests, "value", "values")}…')

	def _explorerSearchesEnd(self):
		'''
		The search has ended.
		'''

		self._explorer_searches = None
		self._explorer_searches_n_tests = None

		self._updateMapperState('Searches ended')

	def _explorerSearchStart(self, search):
		'''
		An Explorer starts to search optimal settings for a particular test.

		Parameters
		----------
		search : dict
			Description of the starting search.
		'''

		self._explorer_search = search

		self._updateMapperState(f'Search {len(self._explorer_searches["searches"])+1}/{self._explorer_searches_n_tests}…')

	def _explorerSearchEnd(self):
		'''
		The search for one test has ended.
		'''

		self._explorer_search = None

		self._updateMapperState('Search ended')

	def _explorerSearchIterationStart(self):
		'''
		An iteration in a search has started.
		'''

		state = f'Search {len(self._explorer_searches["searches"])+1}/{self._explorer_searches_n_tests}'
		state += f', iteration {len(self._explorer_search["iterations"])+1}'

		if self._explorer_search['iterations']:
			state += f', current criterion: {self._explorer_search["iterations"][-1]["stopping_criterion"]}'

		self._updateMapperState(state)

	def _explorerSearchIterationEnd(self):
		'''
		An iteration in a search has ended.
		'''

		pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy

class SimulationBaseSetting(abc.ABC):
	'''
	Represent a setting, either global or normal (abstract class).

	Parameters
	----------
	simulation : Simulation
		The simulation using this setting.

	setting_name : str
		Name of the setting.

	setting_value : mixed
		Default value of the setting.
	'''

	def __init__(self, simulation, setting_name, setting_value):
		self._simulation = simulation

		self._name = setting_name
		self._value = setting_value

	def __deepcopy__(self, memo):
		'''
		Override the default behavior of `deepcopy()` to keep the references to the Simulation.
		'''

		cls = self.__class__
		result = cls.__new__(cls)
		memo[id(self)] = result

		for k, v in self.__dict__.items():
			if k == '_simulation':
				setattr(result, k, v)

			else:
				setattr(result, k, copy.deepcopy(v, memo))

		return result

	@property
	def name(self):
		'''
		Getter for the setting name.

		Returns
		-------
		name : str
			Name of the setting.
		'''

		return self._name

	@property
	def value(self):
		'''
		Getter for the setting value, with fixers applied.

		Returns
		-------
		value : mixed
			Value of the setting.
		'''

		try:
			fixers = self._fixers

		except AttributeError:
			fixers = {}

		return self._simulation.folder.applyFixers(self._simulation.parseString(self._value), **fixers)

	@value.setter
	def value(self, new_value):
		'''
		Setter for the setting value.

		Parameters
		----------
		new_value : mixed
			New value of the setting
		'''

		self._value = new_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil

class LocalSFTP():
	'''
	Implement some actions on local files, with the same methods names than paramiko.SFTP.

	Parameters
	----------
	wd : str
		Working directory (base directory to use for the files).
	'''

	def __init__(self, wd = '.'):
		self.chdir(wd)

	def chdir(self, wd):
		'''
		Change the working directory.
		If the new working directory does not exist, create it.

		Parameters
		----------
		wd : str
			Working directory to use.
		'''

		if not(os.path.isdir(wd)):
			os.makedirs(wd)

		self._wd = wd

	def path(self, path):
		'''
		Prepend a path with the working directory.

		Parameters
		----------
		path : str
			Path to prepend.

		Returns
		-------
		complete_path : str
			The complete path, prepended.
		'''

		return os.path.join(self._wd, path)

	def put(self, local_path, remote_path, replace = False, delete = False):
		'''
		Copy or move a file or folder into the working directory.

		Parameters
		----------
		local_path : str
			Path to the file/folder to copy.

		remote_path : str
			Path of the copied file/folder.

		replace : bool
			If `False`, send a file only if the source is more recent. Otherwise always send it.

		delete : bool
			If `True`, move the local file/folder, if `False`, copy it.
		'''

		send = shutil.move

		if os.path.isfile(local_path):
			if not(replace):
				try:
					if os.stat(local_path).st_mtime <= os.stat(self.path(remote_path)).st_mtime:
						return None

				except FileNotFoundError:
					pass

			if not(delete):
				send = shutil.copy

		else:
			if not(delete):
				send = shutil.copytree

			if replace and os.path.isdir(self.path(remote_path)):
				shutil.rmtree(self.path(remote_path))

		try:
			send(local_path, self.path(remote_path))

		except FileNotFoundError:
			os.makedirs(os.path.dirname(self.path(remote_path)))
			send(local_path, self.path(remote_path))

	def get(self, remote_path, local_path, delete = False):
		'''
		Copy or move a file or folder from the working directory.

		Parameters
		----------
		remote_path : str
			Path to the file/folder to copy.

		local_path : str
			Path of the copied file/folder.

		delete : bool
			If `True`, move the remote file/folder, if `False`, copy it.
		'''

		if os.path.isdir(local_path):
			shutil.rmtree(local_path)

		receive = shutil.move

		if not(delete):
			receive = shutil.copy if os.path.isfile(self.path(remote_path)) else shutil.copytree

		try:
			receive(self.path(remote_path), local_path)

		except FileNotFoundError:
			os.makedirs(os.path.dirname(local_path))
			receive(self.path(remote_path), local_path)

	def remove(self, path):
		'''
		Remove a file or a folder.

		Parameters
		----------
		path : str
			Path to the file/folder to remove.
		'''

		path = self.path(path)

		if os.path.isfile(path):
			os.unlink(path)

		else:
			shutil.rmtree(path)

	def stat(self, path):
		'''
		Get some numbers about a file.

		Parameters
		----------
		path : str
			Path of the file.

		Returns
		-------
		res : os.stat_result
			Result of the stat function.
		'''

		return os.stat(self.path(path))

	def open(self, path, mode):
		'''
		Open a file.

		Parameters
		----------
		path : str
			Path to the file to open.

		mode : str
			Opening mode.

		Returns
		-------
		fid : TextIOWrapper
			Opened file.
		'''

		return open(self.path(path), mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class ManagerError(Exception):
	'''
	Base class for exceptions occurring in the Manager.
	'''

	pass

class ManagerAlreadyRunningError(ManagerError):
	'''
	Exception raised when a instance of the Manager is created while another is still running.
	'''

	pass

class ManagerOperationNotAllowed(ManagerError):
	'''
	Exception raised when we try to perform an operation that is not allowed (e.g. write something in read only mode).
	'''

	pass

class SimulationFolderAlreadyExistError(ManagerError):
	'''
	Exception raised when the folder of a simulation already exists.

	Parameters
	----------
	folder : str
		The name of the folder which should not exist.
	'''

	def __init__(self, folder):
		self.folder = folder

class SimulationFolderNotFoundError(ManagerError):
	'''
	Exception raised when we try to add a simulation to the manager, but the indicated folder does not exist.

	Parameters
	----------
	folder : str
		The folder which has not been found.
	'''

	def __init__(self, folder):
		self.folder = folder

class SimulationIntegrityCheckFailedError(ManagerError):
	'''
	Exception raised when a folder fails to pass an integrity check.

	Parameters
	----------
	folder : str
		The folder which has not been found.
	'''

	def __init__(self, folder):
		self.folder = folder

class SimulationNotFoundError(ManagerError):
	'''
	Exception raised when we look for a non existing simulation.

	Parameters
	----------
	simulation : str
		The name of the simulation which has not been found.
	'''

	def __init__(self, simulation):
		self.simulation = simulation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import shutil
import stat
from string import Template

from .errors import *
from ..folder import Folder
from ..simulation import Simulation
from ..utils import jsonfiles

class Generator():
	'''
	Generate the scripts to create some simulations.

	Parameters
	----------
	folder : Folder|str
		The folder to manage. Either a `Folder` instance or the path to the folder (used to create a `Folder` instance).
	'''

	def __init__(self, folder):
		self._folder = folder if type(folder) is Folder else Folder(folder)

		self._simulations_to_generate = []

		self._forloop_regex_compiled = None

		self._variables = None

	@property
	def folder(self):
		'''
		Return the `Folder` instance.

		Returns
		-------
		folder : Folder
			The instance used by the generator.
		'''

		return self._folder

	@property
	def _forloop_regex(self):
		'''
		Regex for detecting a `for` loop.

		Returns
		-------
		regex : re.Pattern
			The `for` loop regex.
		'''

		if self._forloop_regex_compiled is None:
			self._forloop_regex_compiled = re.compile('^[ \t]*#{3} FOR (?P<varname>[A-Z0-9_]+) FROM (?P<from>\$?[A-Z0-9_]+) TO (?P<to>\$?[A-Z0-9_]+)$.+?^(?P<content>.+?)^[ \t]*#{3}$.+?^', flags = re.MULTILINE | re.DOTALL)

		return self._forloop_regex_compiled

	@property
	def variables(self):
		'''
		Get the latest generated variables.

		Returns
		-------
		variables : dict
			The variables generated for the latest script.
		'''

		return self._variables

	def add(self, simulation):
		'''
		Add a simulation to generate.

		Parameters
		----------
		simulation : list|dict|Simulation
			The simulation(s) to add.
		'''

		if type(simulation) is list:
			for s in simulation:
				self.add(s)

		else:
			self._simulations_to_generate.append(Simulation.ensureType(simulation, self._folder))

	def clear(self):
		'''
		Clear the list of simulations to generate.
		'''

		self._simulations_to_generate.clear()
		self._variables = None

	@property
	def command_lines(self):
		'''
		Get the list of the command lines corresponding to the set of simulations.

		Returns
		-------
		command_lines : list
			The list of command lines.
		'''

		return [simulation.command_line for simulation in self._simulations_to_generate]

	def _createDestinationFolder(self, dest_folder, *, empty_dest = False):
		'''
		Create the folder where the scripts will be stored.

		Parameters
		----------
		dest_folder : str
			Destination folder where scripts should be stored.

		empty_dest : boolean
			If `True` and if the destination folder already exists, empty it before generating the scripts. If `False` the existence of the folder raises an error.

		Raises
		------
		DestinationFolderExistsError
			The destination folder already exists.
		'''

		if os.path.isdir(dest_folder):
			if empty_dest:
				for entry in [os.path.join(dest_folder, e) for e in os.listdir(dest_folder)]:
					(shutil.rmtree if os.path.isdir(entry) else os.unlink)(entry)

			else:
				raise DestinationFolderExistsError()

		else:
			os.makedirs(dest_folder)

	def _exportCommandLines(self, filename):
		'''
		Export the command lines to a JSON file.

		Parameters
		----------
		filename : str
			Name of the file to create.
		'''

		with open(filename, 'w') as f:
			f.write('\n'.join(self.command_lines) + '\n')

	def _loadConfig(self, config_name, basedir):
		'''
		Load the config and set the variables.

		Parameters
		----------
		config_name : str
			Name of the config to load.

		basedir : str
			Base directory to use in the paths.
		'''

		self._config = self._folder.config('generator', config_name)

		self._variables = {
			key.upper(): value
			for key, value in self._config.items()
		}

		self._variables['N_EXEC'] = min(self._variables['N_EXEC'], len(self._simulations_to_generate))

		self._variables['HATENO'] = self._folder.config('folder', config_name)['hateno']

		self._variables['BASEDIR'] = basedir

		self._variables['COMMAND_LINES_FILENAME'] = os.path.join(basedir, 'command_lines.txt')
		self._variables['JOB_DIRECTORY'] = os.path.join(basedir, self._variables['JOB_DIRECTORY'])
		self._variables['LOG_FILENAME'] = os.path.join(basedir, self._variables['LOG_FILENAME'])

		for skeleton_filename in self._folder.skeletons(self._config['skeletons']):
			skeleton = os.path.basename(skeleton_filename)
			varname = skeleton.upper().replace('.', '_')
			self._variables[f'FILE_{varname}'] = os.path.join(basedir, skeleton)

	def _replaceVariables(self, s):
		'''
		Parse a string to replace the possible variables.

		Parameters
		----------
		s : str
			The string to parse.

		Returns
		-------
		parsed : str
			The parsed string.
		'''

		return Template(s).safe_substitute(**self._variables)

	def _replaceForLoop(self, match):
		'''
		Replace the a `for` loop of a skeleton.
		To be called by `re.sub()`.

		Parameters
		----------
		match : re.Match
			Match object from the `for` loop regex.

		Returns
		-------
		loop_content : str
			The loop replaced.
		'''

		loop_from = int(self._replaceVariables(match.group('from')))
		loop_to = int(self._replaceVariables(match.group('to')))

		return ''.join([
			self._replaceVariables(match.group('content'))
			for k in range(loop_from, loop_to + 1)
		])

	def _generateScript(self, skeleton_filename, script_filename):
		'''
		Generate a script from a skeleton.

		Parameters
		----------
		skeleton_filename : str
			Path to the skeleton.

		script_filename : str
			Path to the script to write.
		'''

		with open(skeleton_filename, 'r') as f:
			skeleton = f.read()

		script_content = self._forloop_regex.sub(self._replaceForLoop, skeleton)
		script_content = self._replaceVariables(script_content)

		with open(script_filename, 'w') as f:
			f.write(script_content)

		os.chmod(script_filename, os.stat(script_filename).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

	def generate(self, dest_folder, config_name = None, *, empty_dest = False, basedir = None):
		'''
		Generate the scripts to launch the simulations.

		Parameters
		----------
		dest_folder : str
			Destination folder where scripts should be stored.

		config_name : str
			Name of the config to use.

		empty_dest : boolean
			If `True` and if the destination folder already exists, empty it before generating the scripts. If `False` the existence of the folder raises an error.

		basedir : str
			Path to the "final" directory from which the scripts will be executed.

		Raises
		------
		GeneratorEmptyListError
			The list of simulations to generate is empty.

		Returns
		-------
		script_path : str
			Remote path of the generated script.
		'''

		if not(self._simulations_to_generate):
			raise GeneratorEmptyListError()

		self._createDestinationFolder(dest_folder, empty_dest = empty_dest)
		self._loadConfig(config_name, basedir or dest_folder)

		self._exportCommandLines(os.path.join(dest_folder, 'command_lines.txt'))

		for skeleton_filename in self._folder.skeletons(self._config['skeletons']):
			self._generateScript(skeleton_filename, os.path.join(dest_folder, os.path.basename(skeleton_filename)))

		return os.path.join(self._variables['BASEDIR'], 'exec.sh')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import watchdog.observers.polling

class FileWatcher():
	'''
	Monitor files or directories.

	Parameters
	----------
	event_handler : watchdog.events.FileSystemEventHandler
		Object used to handle the events.

	path : str
		Path to the file or directory to monitor.

	timeout : float
		Interval (in seconds) between polling the file system.
	'''

	def __init__(self, event_handler, path, *, timeout = 0.1):
		self._observer = watchdog.observers.polling.PollingObserver(timeout = timeout)
		self._observer.schedule(event_handler, path)

	def __enter__(self):
		'''
		Context manager to call `start()` and `stop()` automatically.
		'''

		self.start()

	def __exit__(self, *args, **kwargs):
		'''
		Ensure `stop()` is called when exiting the context manager.
		'''

		self.stop()

	def start(self):
		'''
		Start the monitoring.
		'''

		self._observer.start()

	def stop(self):
		'''
		Stop the monitoring.
		'''

		self._observer.stop()
		self._observer.join()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class GeneratorError(Exception):
	'''
	Base class for exceptions occurring in the Generator.
	'''

	pass

class DestinationFolderExistsError(GeneratorError):
	'''
	Exception raised when a destination folder already exists.
	'''

	pass

class GeneratorEmptyListError(GeneratorError):
	'''
	Exception raised when a specific list is empty.
	'''

	pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class RemoteFolderError(Exception):
	'''
	Base class for exceptions occurring in a RemoteFolder.
	'''

	pass

class RemotePathNotFoundError(RemoteFolderError):
	'''
	Exception raised when we try to access a remote path/directory which does not exist.

	Parameters
	----------
	remote_path : str
		The path.
	'''

	def __init__(self, remote_path):
		self.remote_path = remote_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import errno
import glob
import os
import tempfile

from . import checkers as default_checkers
from . import fixers as default_fixers
from . import namers as default_namers
from .errors import *
from ..utils import FCollection, utils, string, jsonfiles

MAIN_FOLDER = '.hateno'
CONFIG_FOLDER = 'config'
SKELETONS_FOLDER = 'skeletons'
SIMULATIONS_FOLDER = 'simulations'
TMP_FOLDER = 'tmp'

CONF_FILENAME = 'hateno.conf'
SIMULATIONS_LIST_FILENAME = 'simulations.list'
RUNNING_MANAGER_INDICATOR_FILENAME = 'manager.running'

class Folder():
	'''
	Base class for each system needing access to the configuration files of a simulations folder.
	Initialize with the simulations folder and load the settings.

	Parameters
	----------
	folder : str
		The simulations folder. Must contain a settings file.

	Raises
	------
	FileNotFoundError
		No configuration file found in the configuration folder.
	'''

	def __init__(self, folder):
		self._folder = folder
		self._conf_folder_path = os.path.join(self._folder, MAIN_FOLDER)
		self._settings_file = os.path.join(self._conf_folder_path, CONF_FILENAME)
		self._tmp_dir = os.path.join(self._conf_folder_path, TMP_FOLDER)

		if not(os.path.isfile(self._settings_file)):
			raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self._settings_file)

		if not(os.path.isdir(self._tmp_dir)):
			os.makedirs(self._tmp_dir)

		self._settings = None

		self._config_folders_dict = None
		self._configs = {}

		self._skeletons_folders_dict = None
		self._skeletons = {}

		self._program_files = None

		self._namers = None
		self._fixers = None
		self._checkers = None
		self._evaluations = None

	@property
	def folder(self):
		'''
		Return the folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._folder

	@property
	def tmp_folder(self):
		'''
		Return the temporary folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._tmp_dir

	def tempdir(self):
		'''
		Create a temporary directory.

		Returns
		-------
		path : str
			The path to the created folder.
		'''

		return tempfile.mkdtemp(dir = self._tmp_dir)

	def _relpath(self, path):
		'''
		Transform a path relative to the configuration folder.

		Parameters
		----------
		path : str
			Path (relative to the configuration folder) to transform.

		Returns
		-------
		relative_path : str
			The path.
		'''

		return os.path.normpath(os.path.join(self._conf_folder_path, path))

	@property
	def _config_folders(self):
		'''
		The list of the available configuration folders.

		Returns
		-------
		folders : dict
			A dictionary associating the folders' names to their paths.
		'''

		if self._config_folders_dict is None:
			self._config_folders_dict = {}

			# First, the imported folders, then the local one.
			# In this way, the local configs will always overwrite the imported ones.

			for import_desc in self.settings['import']:
				try:
					to_import = import_desc['config']

				except KeyError:
					pass

				else:
					self._config_folders_dict.update({
						foldername: self._relpath(os.path.join(import_desc['from'], MAIN_FOLDER, CONFIG_FOLDER, foldername))
						for foldername in to_import
					})

			base_folder = os.path.join(self._conf_folder_path, CONFIG_FOLDER)

			if os.path.isdir(base_folder):
				for foldername in os.listdir(base_folder):
					path = os.path.join(base_folder, foldername)

					if os.path.isdir(path):
						self._config_folders_dict[foldername] = path

		return self._config_folders_dict

	def configname(self, foldername = None):
		'''
		Get the name of a configuration.

		Parameters
		----------
		foldername : str
			Name of the configuration folder. If `None`, use the default config indicated in the configuration file.

		Raises
		------
		NoConfigError
			No configuration folder name given.

		ConfigNotFoundError
			The asked config name is not defined.

		Returns
		-------
		foldername : str
			Name of the configuration folder.
		'''

		foldername = foldername or self.settings.get('default_config')

		if foldername is None:
			if len(self._config_folders) != 1:
				raise NoConfigError()

			else:
				foldername = list(self._config_folders.keys())[0]

		if foldername not in self._config_folders:
			raise ConfigNotFoundError(foldername)

		return foldername

	def configFilepath(self, filename, foldername = None):
		'''
		Get the path to a file in a config folder.

		Parameters
		----------
		filename : str
			Name of the file.

		foldername : str
			Name of the configuration folder. If `None`, use the default config indicated in the configuration file.

		Raises
		------
		NoConfigError
			No configuration folder name given.

		Returns
		-------
		path : str
			Path to the file.
		'''

		foldername = self.configname(foldername)
		return os.path.join(self._config_folders[foldername], filename)

	def config(self, configname, foldername = None):
		'''
		Get a configuration object.

		Parameters
		----------
		configname : str
			Name of the wanted configuration.

		foldername : str
			Name of the configuration folder. If `None`, use the default config indicated in the configuration file.

		Returns
		-------
		config : dict
			Dictionary stored in the right configuration file.
		'''

		foldername = self.configname(foldername)

		if foldername not in self._configs:
			self._configs[foldername] = {}

		if configname not in self._configs[foldername]:
			try:
				self._configs[foldername][configname] = jsonfiles.read(self.configFilepath(f'{configname}.json', foldername))

			except FileNotFoundError:
				self._configs[foldername][configname] = None

		return self._configs[foldername][configname]

	@property
	def _skeletons_folders(self):
		'''
		The list of the available skeletons.

		Returns
		-------
		folders : dict
			A dictionary associating the folders' names to their paths.
		'''

		if self._skeletons_folders_dict is None:
			self._skeletons_folders_dict = {}

			for import_desc in self.settings['import']:
				try:
					to_import = import_desc['skeletons']

				except KeyError:
					pass

				else:
					self._skeletons_folders_dict.update({
						foldername: self._relpath(os.path.join(import_desc['from'], MAIN_FOLDER, SKELETONS_FOLDER, foldername))
						for foldername in to_import
					})

			base_folder = os.path.join(self._conf_folder_path, SKELETONS_FOLDER)

			if os.path.isdir(base_folder):
				for foldername in os.listdir(base_folder):
					path = os.path.join(base_folder, foldername)

					if os.path.isdir(path):
						self._skeletons_folders_dict[foldername] = path

		return self._skeletons_folders_dict

	def skeletons(self, skeletons_name):
		'''
		Get the skeletons filepaths.

		Parameters
		----------
		skeletons_name : str
			Name of the skeletons folder.

		Raises
		------
		SkeletonsNotFoundError
			The skeletons folder does not exist.

		Returns
		-------
		filepaths : list
			The paths of the skeletons.
		'''

		if skeletons_name not in self._skeletons_folders:
			raise SkeletonsNotFoundError(skeletons_name)

		if skeletons_name not in self._skeletons:
			path = self._skeletons_folders[skeletons_name]
			self._skeletons[skeletons_name] = [os.path.join(path, filename) for filename in os.listdir(path)]

		return self._skeletons[skeletons_name]

	@property
	def simulations_list_filename(self):
		'''
		Return the path to the file where the list of simulations is stored.

		Returns
		-------
		path : str
			Path to the simulations list file.
		'''

		return os.path.join(self._conf_folder_path, SIMULATIONS_LIST_FILENAME)

	@property
	def simulations_folder(self):
		'''
		Return the path to the folder where the simulations are stored.
		Create the folder if it does not exist.

		Returns
		-------
		path : str
			Path to the simulations folder.
		'''

		path = os.path.join(self._conf_folder_path, SIMULATIONS_FOLDER)
		if not(os.path.isdir(path)):
			os.makedirs(path)

		return path

	@property
	def running_manager_indicator_filename(self):
		'''
		Return the path to the file indicating the Manager is currently running.

		Returns
		-------
		path : str
			Path to the indicator file.
		'''

		return os.path.join(self._conf_folder_path, RUNNING_MANAGER_INDICATOR_FILENAME)

	@property
	def settings(self):
		'''
		Return the content of the settings file as a dictionary.

		Returns
		-------
		settings : dict
			The folder's settings.
		'''

		if self._settings is None:
			self._settings = jsonfiles.read(self._settings_file)

			if 'import' not in self._settings:
				self._settings['import'] = []

			elif type(self._settings['import']) is not list:
				self._settings['import'] = [self._settings['import']]

			if 'namers' not in self._settings:
				self._settings['namers'] = []

			if 'fixers' not in self._settings:
				self._settings['fixers'] = []

		return self._settings

	@property
	def program_files(self):
		'''
		Get the list of the files defined in the configuration file.

		Returns
		-------
		files : list
			A list of files. Each item is a tuple. First item is the local path, second item is the remote one.
		'''

		if self._program_files is None:
			self._program_files = []

			try:
				for path_item in self.settings['files']:
					given_path, dest = path_item if type(path_item) is list else (path_item, '')

					for path in glob.glob(os.path.normpath(os.path.join(self._conf_folder_path, given_path))):
						if os.path.isfile(path):
							self._program_files.append((path, os.path.join(dest, os.path.basename(path))))

						else:
							self._program_files += [
								(
									os.path.join(root, file),
									os.path.join(dest, os.path.relpath(os.path.join(root, file), os.path.join(path, '..')))
								)
								for root, folders, files in os.walk(path)
								for file in files
							]

			except KeyError:
				pass

		return self._program_files

	def _loadFCollection(self, filter_regex, default_module, custom_filename, categories = []):
		'''
		Load a collection of functions used in the folder (fixers, namers, etc.).
		First load the default functions, then the imported ones, and finally the custom ones.
		In that way, the custom functions, defined in the current folder, will always overwrite the default and imported ones.

		Parameters
		----------
		filter_regex : str
			Regex to use to filter the functions in a module.

		default_module : module
			Module where the default functions are defined.

		custom_filename : str
			Name of the file (without extension) where the custom functions should be defined.

		categories : list
			List of categories to use in the FCollection.

		Returns
		-------
		collection : FCollection
			The loaded collection.
		'''

		collection = FCollection(categories = categories, filter_regex = filter_regex)

		if default_module is not None:
			collection.loadFromModule(default_module)

		for import_desc in self.settings['import']:
			try:
				to_import = import_desc['functions']

			except KeyError:
				pass

			else:
				if custom_filename in to_import:
					import_file = self._relpath(os.path.join(import_desc['from'], MAIN_FOLDER, f'{custom_filename}.py'))

					if os.path.isfile(import_file):
						collection.loadFromModule(utils.loadModuleFromFile(import_file))

		custom_file = os.path.join(self._conf_folder_path, f'{custom_filename}.py')
		if os.path.isfile(custom_file):
			collection.loadFromModule(utils.loadModuleFromFile(custom_file))

		return collection

	@property
	def fixers(self):
		'''
		Get the list of available values fixers.

		Returns
		-------
		fixers : FCollection
			The collection of values fixers.
		'''

		if self._fixers is None:
			self._fixers = self._loadFCollection(r'^fixer_(?P<name>[A-Za-z0-9_]+)$', default_fixers, 'fixers')

		return self._fixers

	@property
	def namers(self):
		'''
		Get the list of available namers.

		Returns
		-------
		namers : FCollection
			The collection of namers.
		'''

		if self._namers is None:
			self._namers = self._loadFCollection(r'^namer_(?P<name>[A-Za-z0-9_]+)$', default_namers, 'namers')

		return self._namers

	@property
	def checkers(self):
		'''
		Get the list of available checkers.

		Returns
		-------
		checkers : FCollection
			The collection of checkers.
		'''

		if self._checkers is None:
			self._checkers = self._loadFCollection(r'^(?P<category>file|folder|global)_(?P<name>[A-Za-z0-9_]+)$', default_checkers, 'checkers', ['file', 'folder', 'global'])

		return self._checkers

	@property
	def evaluations(self):
		'''
		Get the list of available evaluation functions.

		Returns
		-------
		evaluations : FCollection
			The collection of evaluation functions.
		'''

		if self._evaluations is None:
			self._evaluations = self._loadFCollection(r'^eval_(?P<name>[A-Za-z0-9_]+)$', None, 'evaluations')

		return self._evaluations

	def applyFixers(self, value, *, before = [], after = []):
		'''
		Fix a value to prevent false duplicates (e.g. this prevents to consider `0.0` and `0` as different values).
		Each item of a list of fixers is either a fixer's name or a list beginning with the fixer's name and followed by the arguments to pass to the fixer.

		Parameters
		----------
		value : mixed
			The value to fix.

		before : list
			List of fixers to apply before the global ones.

		after : list
			List of fixers to apply after the global ones.

		Returns
		-------
		fixed : mixed
			The same value, fixed.
		'''

		value = copy.deepcopy(value)

		for fixer in before + self.settings['fixers'] + after:
			value = self.fixers.call(fixer, value)

		return value

	def applyNamers(self, setting, *, before = [], after = []):
		'''
		Transform the name of a setting before being used in a simulation.

		Parameters
		----------
		setting : dict
			Representation of the setting.

		before : list
			List of namers to apply before the global ones.

		after : list
			List of namers to apply after the global ones.

		Returns
		-------
		name : str
			The name to use.
		'''

		for namer in before + self.settings['namers'] + after:
			setting['name'] = self.namers.call(namer, setting)

		return setting['name']

	def checkIntegrity(self, simulation):
		'''
		Check the integrity of a simulation.

		Parameters
		----------
		simulation : Simulation
			The simulation to check.

		Returns
		-------
		success : bool
			`True` if the integrity check is successful, `False` otherwise.
		'''

		tree = {}

		for output_entry in ['files', 'folders']:
			tree[output_entry] = []
			checkers_cat = output_entry[:-1]

			if output_entry in self.settings['output']:
				for output in self.settings['output'][output_entry]:
					parsed_name = str(simulation.parseString(output['name']))
					tree[output_entry].append(parsed_name)

					if 'checks' in output:
						for checker in output['checks']:
							if not(self.checkers.call(checker, simulation, parsed_name, category = checkers_cat)):
								return False

		if 'checks' in self.settings['output']:
			for checker in self.settings['output']['checks']:
				if not(self.checkers.call(checker, simulation, tree, category = 'global')):
					return False

		return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import re
import os

from .globalsetting import SimulationGlobalSetting
from .setting import SimulationSetting
from .errors import *
from ..utils import string, jsonfiles

class Simulation():
	'''
	Represent a simulation, itself identified by its settings.

	Parameters
	----------
	folder : Folder
		Instance of `Folder` for the current simulation's folder.

	settings : dict
		Dictionary listing the user settings.
	'''

	def __init__(self, folder, settings):
		self._folder = folder
		self._user_settings = settings

		self._raw_globalsettings = None
		self._raw_settings = None

		self._indexed_settings = None

		self._setting_tag_regex_compiled = None
		self._eval_tag_regex_compiled = None
		self._parser_recursion_stack = []

	@classmethod
	def ensureType(cls, simulation, folder):
		'''
		Ensure a variable is a Simulation object.

		Parameters
		----------
		simulation : dict|Simulation
			The variable to check.

		folder : Folder
			The `Folder` instance to use in case we need to create a new object.

		Returns
		-------
		simulation : Simulation
			The simulation as a Simulation instance.
		'''

		if type(simulation) is cls:
			return simulation

		return cls(folder, simulation)

	def copy(self):
		'''
		Create a copy of the simulation.

		Returns
		-------
		simulation : Simulation
			The copy of the simulation.
		'''

		return Simulation(self._folder, {
			**self.globalsettings,
			'settings': {
				settings_set_name: [
					{s.name: s._value for s in settings_set}
					for settings_set in settings_sets
				]
				for settings_set_name, settings_sets in self._settings.items()
			}
		})

	@property
	def folder(self):
		'''
		Get the Folder instance used by the simulation.

		Returns
		-------
		folder : Folder
			The Folder instance.
		'''

		return self._folder

	def __getitem__(self, key):
		'''
		Access to a global setting.

		Parameters
		----------
		key : str
			The key of the setting to get.

		Raises
		------
		KeyError
			The key does not exist.

		Returns
		-------
		value : mixed
			The corresponding value.
		'''

		try:
			return self.globalsettings[key]

		except KeyError:
			raise KeyError('The key does not exist in the global settings')

	def __setitem__(self, key, value):
		'''
		Change a global setting.

		Parameters
		----------
		key : str
			The key of the setting to change.

		value : mixed
			The new value of the setting.

		Raises
		------
		KeyError
			The key does not exist.
		'''

		try:
			setting = [setting for setting in self._globalsettings if setting.name == key][0]

		except IndexError:
			raise KeyError('The key does not exist in the global settings')

		else:
			setting.value = value

	def writeSettingsFile(self, filename, *, folder = None):
		'''
		Write the settings into a file in the simulation's folder.

		Parameters
		----------
		filename : str
			Name of the file to write.

		folder : str
			Name of the folder in which the file should be saved. Default to the simulation's folder.
		'''

		jsonfiles.write(self.settings, os.path.join(folder or self['folder'], filename))

	@property
	def _globalsettings(self):
		'''
		Return (and generate if needed) the complete list of global settings.

		Returns
		-------
		raw_globalsettings : list
			The global settings.
		'''

		if self._raw_globalsettings is None:
			self.generateGlobalSettings()

		return self._raw_globalsettings

	@property
	def _settings(self):
		'''
		Return (and generate if needed) the complete list of settings as a dictionary.

		Returns
		-------
		raw_settings : dict
			The settings.
		'''

		if self._raw_settings is None:
			self.generateSettings()

		return self._raw_settings

	@property
	def raw_settings(self):
		'''
		Return a dictionary with the complete list of sets of settings.

		Returns
		-------
		settings : dict
			The list of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s for s in settings_set}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings.items()
		}

	def getSetting(self, coords):
		'''
		Retrieve a setting from its set and name.

		Parameters
		----------
		coords : dict
			Describe the "coordinates" of the setting. Must contain the following keys:
				* `set`: the name of the set the setting belongs to,
				* `name`: the name of the setting.
			The following key is optional:
				* `set_index`: the index of the set. If not provided, default to 0.

		Returns
		-------
		setting : SimulationSetting
			The setting corresponding to the coordinates.
		'''

		coords = {
			'set_index': 0,
			**coords
		}

		return self.raw_settings[coords['set']][coords['set_index']][coords['name']]

	@property
	def settings(self):
		'''
		Return a dictionary with the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : dict
			List of sets of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s.value for s in settings_set if not(s.exclude)}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings.items()
		}

	@property
	def raw_values_settings(self):
		'''
		Return a dictionary with the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : dict
			List of sets of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s._value for s in settings_set}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings.items()
		}

	@property
	def settings_as_strings(self):
		'''
		Return the complete list of sets of settings to use, as strings.
		Take into account the `only_if` parameter.

		Returns
		-------
		settings : dict
			Settings, generated according to their pattern and organized by sets.
		'''

		return {
			set_name: [
				[str(s) for s in settings_set if s.shouldBeDisplayed()]
				for settings_set in settings_sets
			]
			for set_name, settings_sets in self._settings.items()
		}

	@property
	def globalsettings(self):
		'''
		Return the list of global settings, as a name: value dictionary.

		Returns
		-------
		settings : dict
			The global settings.
		'''

		return {setting.name: setting.value for setting in self._globalsettings}

	@property
	def command_line(self):
		'''
		Return the command line to use to generate this simulation.

		Returns
		-------
		command_line : str
			The command line to execute.
		'''

		return ' '.join([self._folder.settings['exec']] + sum(sum(self.settings_as_strings.values(), []), []))

	@property
	def _setting_tag_regex(self):
		'''
		Regex to detect whether there is a setting or global setting tag in a string.

		Returns
		-------
		regex : re.Pattern
			The setting tag regex.
		'''

		if self._setting_tag_regex_compiled is None:
			self._setting_tag_regex_compiled = re.compile(r'\{(?P<category>(?:global)?setting):(?:(?P<setname>.+?)(?:\[(?P<index>[0-9]+)\])?\.)?(?P<name>.+?)\}')

		return self._setting_tag_regex_compiled

	@property
	def _eval_tag_regex(self):
		'''
		Regex to detect whether we need to evaluate a part of a string.

		Returns
		-------
		regex : re.Pattern
			The eval tag regex.
		'''

		# The `\)*` part is needed so we don't have troubles with functions calls right before the end of a string.
		# Without this, such calls (e.g. in "((sqrt(16)))") end the string prematurely.

		if self._eval_tag_regex_compiled is None:
			self._eval_tag_regex_compiled = re.compile(r'\(\((.*?\)*)\)\)')

		return self._eval_tag_regex_compiled

	def generateGlobalSettings(self):
		'''
		Generate the full list of global settings.
		'''

		self._raw_globalsettings = []

		for setting in self._folder.settings['globalsettings']:
			try:
				setting_value = self._user_settings[setting['name']]

			except KeyError:
				setting_value = setting['default']

			self._raw_globalsettings.append(SimulationGlobalSetting(self, setting['name'], setting_value))

	def getSettingCount(self, setting_name, set_name = None):
		'''
		Get the total number of uses of a given setting, globally or locally.

		Parameters
		----------
		setting_name : str
			Name of the setting to get the count of.

		set_name : str
			Name of the set for the local count.

		Raises
		------
		SettingsSetNotFoundError
			The provided set name does not exist.

		SettingNotFoundError
			The provided setting name does not exist.

		Returns
		-------
		count : int
			Local count if `set_name` is not `None`, global count otherwise.
		'''

		try:
			settings = self._indexed_settings['global'] if set_name is None else self._indexed_settings['local'][set_name]

		except KeyError:
			raise SettingsSetNotFoundError(set_name)

		else:
			try:
				return len(settings[setting_name])

			except KeyError:
				raise SettingNotFoundError(set_name, setting_name)

	def _indexSetting(self, setting, set_name):
		'''
		Add a setting to the index.

		Parameters
		----------
		setting : SimulationSetting
			The setting to index.

		set_name : str
			The name of the set the setting belongs to.
		'''

		if not(set_name in self._indexed_settings['local']):
			self._indexed_settings['local'][set_name] = {}

		indexes = []

		for indexes_dict in [self._indexed_settings['global'], self._indexed_settings['local'][set_name]]:
			try:
				indexes_dict[setting.name].append(setting)

			except KeyError:
				indexes_dict[setting.name] = [setting]

			indexes.append(len(indexes_dict[setting.name]) - 1)

		setting.setIndexes(*indexes)

	def _addRawSettingsSet(self, set_name, default_settings, values_set = {}):
		'''
		Add a raw settings set to the list.

		Parameters
		----------
		set_name : str
			Name of the set.

		default_settings : list
			List of the default settings of this set.

		values_set : dict
			Values to override the defaults.
		'''

		set_to_add = copy.deepcopy(default_settings)

		for setting in set_to_add:
			self._indexSetting(setting, set_name)

			try:
				setting.value = values_set[setting.name]

			except KeyError:
				pass

		try:
			self._raw_settings[set_name].append(set_to_add)

		except KeyError:
			self._raw_settings[set_name] = [set_to_add]

	def generateSettings(self):
		'''
		Generate the full list of settings, taking into account the user settings and the default values in the folder.
		The "raw settings" are generated.
		Each set of settings is a list of all settings in this set.
		'''

		if type(self._user_settings['settings']) is list:
			user_settings = {
				setting_set_name: [s['settings'] for s in self._user_settings['settings'] if s['set'] == setting_set_name]
				for setting_set_name in set([s['set'] for s in self._user_settings['settings']])
			}

		else:
			user_settings = self._user_settings['settings']

		default_pattern = self._folder.settings['setting_pattern']

		self._raw_settings = {}
		self._indexed_settings = {'global': {}, 'local': {}}

		for settings_set in self._folder.settings['settings']:
			default_settings = [
				SimulationSetting(self, settings_set['set'], s['name'])
				for s in settings_set['settings']
			]

			try:
				values_sets = user_settings[settings_set['set']]

			except KeyError:
				if settings_set['required']:
					self._addRawSettingsSet(settings_set['set'], default_settings)

			else:
				if not(type(values_sets) is list):
					values_sets = [values_sets]

				for values_set in values_sets:
					self._addRawSettingsSet(settings_set['set'], default_settings, values_set)

	def getSettingValueFromTag(self, match):
		'''
		Retrieve the value of a setting from a setting tag.

		Parameters
		----------
		match : re.Match
			Match object corresponding to a setting tag.

		Raises
		------
		SettingTagNotRecognizedError
			The setting tag has not been recognized/does not refer to an existing setting.

		Returns
		-------
		setting : mixed
			The value of the setting.
		'''

		try:
			if match.group('category') == 'globalsetting':
				return [
					setting.value
					for setting in self._globalsettings
					if setting.name == match.group('name')
				][0]

			set_dict = self._indexed_settings['global'] if match.group('setname') is None else self._indexed_settings['local'][match.group('setname')]
			set_list = set_dict[match.group('name')]

			k = 0 if match.group('index') is None else int(match.group('index'))

			return set_list[k].value

		except:
			raise SettingTagNotRecognizedError(match.group(0))

	def replaceSettingTag(self, match):
		'''
		Replace a setting tag (`{setting:set_name[k].setting_name}`) by the value of the right setting.
		To be called by `re.sub()`.

		Parameters
		----------
		match : re.Match
			Match object corresponding to a setting tag.

		Returns
		-------
		setting_value : str
			The value of the setting.
		'''

		try:
			return str(self.getSettingValueFromTag(match))

		except SettingTagNotRecognizedError:
			return match.group(0)

	def parseString(self, s):
		'''
		Parse a string to take into account possible settings.
		The tag `{setting:name}` is replaced by the value of the simulation's setting named `name`.
		The tag `{globalsetting:name}` is replaced by the value of the global setting named `name`.

		Tags are replaced recursively.

		Parameters
		----------
		s : str
			The string to parse.

		Returns
		-------
		parsed : mixed
			The parsed string, or a copy of the setting if the whole string is just one tag.
		'''

		# If the string is not a string, we don't have anything to do (seems reasonable!)
		# Then, we test if the string represents a number, so we can cast it

		if not(type(s) is str):
			return s

		try:
			return float(s)

		except ValueError:
			pass

		s = s.strip()

		# We search for settings tags in the string, and recursively replace them

		if self._raw_settings is None:
			self.generateSettings()

		fullmatch = self._setting_tag_regex.fullmatch(s)

		if fullmatch:
			try:
				return copy.deepcopy(self.getSettingValueFromTag(fullmatch))

			except KeyError:
				return s

		parsed = self._setting_tag_regex.sub(self.replaceSettingTag, s)

		self._parser_recursion_stack.append(s)

		if not(parsed in self._parser_recursion_stack):
			return self.parseString(parsed)

		self._parser_recursion_stack.clear()

		# Final step: we try to evaluate the needed parts of the string to apply allowed operations, if any.
		# ValueError is raised if the string contains any unallowed operation, like the use of exec() or other evil functions.

		try:
			fullmatch = self._eval_tag_regex.fullmatch(parsed)

			if fullmatch:
				parsed = string.safeEval(fullmatch.group(1))

			else:
				parsed = self._eval_tag_regex.sub(lambda m: str(string.safeEval(m.group(1))), parsed)

		except (SyntaxError, ValueError):
			pass

		return parsed



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class FCollectionError(Exception):
	'''
	Base class for exceptions occurring in an FCollection.
	'''

	pass

class FCollectionCategoryNotFoundError(FCollectionError):
	'''
	Exception raised when we try to access a non-existing category of an FCollection.

	Parameters
	----------
	category : str
		Name of the category.
	'''

	def __init__(self, category):
		self.category = category

class FCollectionFunctionNotFoundError(FCollectionError):
	'''
	Exception raised when we try to access a non-existing function of an FCollection.

	Parameters
	----------
	fname : str
		Name of the function.
	'''

	def __init__(self, fname):
		self.fname = fname

class FCollectionInvalidFilterRegexError(FCollectionError):
	'''
	Exception raised when we define the filter regex of an FCollection without the required groups.

	Parameters
	----------
	regex : str
		The invalid regex.
	'''

	def __init__(self, regex):
		self.regex = regex

class EventsError(Exception):
	'''
	Base class for exceptions occurring with the Events class.
	'''

	pass

class EventUnknownError(EventsError):
	'''
	Exception raised when we try to use an unkown event.

	Parameters
	----------
	event : str
		Name of the event.
	'''

	def __init__(self, event):
		self.event = event



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import functools

class UIDisplayedItem(abc.ABC):
	'''
	Represent an item displayed in the UI (abstract class).

	Parameters
	----------
	ui : UI
		The UI object this item belongs to.
	'''

	def __init__(self, ui):
		self.ui = ui
		self.position = self.ui._cursor_vertical_pos

	@abc.abstractproperty
	def height(self):
		'''
		The number of lines used by the item.

		Returns
		-------
		height : int
			The number of lines.
		'''

		pass

	@abc.abstractproperty
	def width(self):
		'''
		The width, in characters, of the item.

		Returns
		-------
		width : int
			The width of the item.
		'''

		pass

	@classmethod
	def renderer(cls, func):
		'''
		Decorator for children's render() method.
		'''

		@functools.wraps(func)
		def wrapper(self, *args, **kwargs):
			self.ui.moveCursorTo(self.position)

			func(self, *args, **kwargs)

			self.ui.moveToLastLine()

		return wrapper

	@abc.abstractmethod
	def render(self):
		'''
		Render the item.
		'''

		pass

	def clear(self):
		'''
		Display enough spaces to clear the object.
		'''

		self.ui.moveCursorTo(self.position)
		print(' ' * self.width, end = '\r')
		self.ui.moveToLastLine()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
import time

import watchdog.events

from .errors import *
from .filewatcher import FileWatcher
from ..utils import Events, jsonfiles

class JobServer():
	'''
	Server part of a job, distributing the command lines over the clients.

	Parameters
	----------
	command_lines_filename : str
		Path to the file listing the command lines to execute (one command line per line).

	job_dir : str
		Path to the job directory to create to store the messaging files.
	'''

	def __init__(self, command_lines_filename, job_dir):
		self._command_lines_filename = command_lines_filename
		self._job_dir = job_dir

		self._event_handler = FileEventHandler(self._job_dir, self._newClient, self._clientGone, self._processRequest)
		self._watcher = FileWatcher(self._event_handler, self._job_dir)

		self._clients = []

		self._log = []

		self.events = Events([
			'client-new', 'client-gone',
			'sent', 'log'
		])

	def __enter__(self):
		'''
		Context manager to automatically call `stop()`.
		'''

		self.start()
		return self

	def __exit__(self, *args, **kwargs):
		'''
		Call `stop()` after exiting a context manager.
		'''

		self.stop()

	def start(self):
		'''
		Start the server.
		Create the job directory.
		Create the generator providing the command lines.

		Raises
		------
		JobDirAlreadyExistsError
			The folder indicated as job directory already exists.
		'''

		try:
			os.makedirs(self._job_dir)

		except FileExistsError:
			raise JobDirAlreadyExistsError(self._job_dir)

		self._command_lines_file = open(self._command_lines_filename, 'r')
		self._command_lines = (row.strip() for row in self._command_lines_file)

		self._watcher.start()

	def stop(self):
		'''
		Stop the server: close the command lines file and remove the job directory.
		'''

		self._watcher.stop()
		self._command_lines_file.close()
		shutil.rmtree(self._job_dir)

	@property
	def log(self):
		'''
		Get the current log.

		Returns
		-------
		log : list
			List of the logs sent by the clients.
		'''

		return self._log

	def _newClient(self, id):
		'''
		A new client has been created: add it to the clients list.

		Parameters
		----------
		id : str
			ID of the client.
		'''

		self._clients.append(id)
		self.events.trigger('client-new', id)

	def _clientGone(self, id):
		'''
		A client has been deleted: remote it from the clients list.

		Parameters
		----------
		id : str
			ID of the client.
		'''

		self._clients.remove(id)
		self.events.trigger('client-gone', id)

	def _processRequest(self, client_id, req):
		'''
		Process the request from a client.

		Parameters
		----------
		client_id : str
			ID of the client.

		req : dict
			Request sent by the client.
		'''

		if 'log' in req:
			self._logCmd(req['log'])

		if req['state'] == 'ready':
			self._sendNextCommandLine(client_id)

	def _sendNextCommandLine(self, client_id):
		'''
		Send the next command line to a client.
		If there is no command line anymore, send `None`.

		Parameters
		----------
		client_id : str
			ID of the client to send to command line to.
		'''

		try:
			cmd = next(self._command_lines)

		except StopIteration:
			cmd = None

		finally:
			jsonfiles.write({
				'command_line': cmd
			}, os.path.join(self._job_dir, f'{client_id}.json'))

			self.events.trigger('sent', cmd, client_id)

	def _logCmd(self, res):
		'''
		Log the result of a command line.

		Parameters
		----------
		res : dict
			Result to log.
		'''

		self._log.append(res)
		self.events.trigger('log', res)

	def run(self):
		'''
		Run loop, stopped when there is no client anymore.
		'''

		while not(self._log) or self._clients:
			try:
				time.sleep(0.1)

			except KeyboardInterrupt:
				break

class FileEventHandler(watchdog.events.RegexMatchingEventHandler):
	'''
	Handle the events from the files.

	Parameters
	----------
	job_dir : str
		Path to the job directory.

	new_client : function
		Function to call when a new client is created.

	client_gone : function
		Function to call when a client is deleted.

	request : function
		Function to call to handle the request from a client.
	'''

	def __init__(self, job_dir, new_client, client_gone, request):
		super().__init__(regexes = [os.path.join(job_dir, r'[0-9a-f]+\.json')])

		self._new_client = new_client
		self._client_gone = client_gone
		self._request = request

		self._prev_message = {}

	def on_created(self, evt):
		'''
		A file has been created in the job directory: a new client is here.
		Call the right callback. We then call `on_modified()` to be sure the first content of the file is processed.

		Parameters
		----------
		evt : watchdog.events.FileCreatedEvent
			Event object from Watchdog containing infos about the client file.
		'''

		client_id = self._filename2id(evt.src_path)
		self._prev_message[client_id] = None

		self._new_client(client_id)

		self.on_modified(evt)

	def on_deleted(self, evt):
		'''
		A file has been deleted: say goodbye to a client.

		Parameters
		----------
		evt : watchdog.events.FileDeletedEvent
			Event object from Watchdog containing infos about the client file.
		'''

		client_id = self._filename2id(evt.src_path)
		del self._prev_message[client_id]

		self._client_gone(client_id)

	def on_modified(self, evt):
		'''
		A file has been modified: maybe a message from a client.
		If it is the case, call the right callback.

		Parameters
		----------
		evt : watchdog.events.FileModifiedEvent
			Event object from Watchdog containing infos about the client file.
		'''

		try:
			content = jsonfiles.read(evt.src_path)

		except:
			pass

		else:
			client_id = self._filename2id(evt.src_path)

			if content == self._prev_message[client_id]:
				return

			self._prev_message[client_id] = content

			if 'state' in content:
				self._request(client_id, content)

	def _filename2id(self, filename):
		'''
		Convert a client filename into an ID.
		It is simply the base name of the file, without extension.

		Parameters
		----------
		filename : str
			Path to the client file.
		'''

		return os.path.splitext(os.path.basename(filename))[0]
